import math
import numbers
import numpy as np


__all__ = ['Vec2', 'Vec3', 'Vec4', 'cross', 'dot', 'interpolate']


class Vector(np.ndarray):
    """Base class for Vec2, Vec3 and Vec4
    Examples:
        >>> v0 = Vec3()  # all zeros
        >>> v1 = Vec3(1.0)  # filled with 1.0
        >>> v2 = Vec3(1.0, 2.0, 3.0) # x, y, z
        >>> v3 = Vec3([1.0, 2.0, 3.0])
        >>> v4 = Vec3((1.0, 2.0, 3.0))
        >>> v5 = Vec3(np.array([1.0, 2.0, 3.0]))
    """
    _shape = None

    def __new__(cls, *args, **kwargs):
        dtype = kwargs.get('dtype', float)

        if cls._shape is not None:
            arg_count = len(args)
            obj = np.zeros((cls._shape[0],), dtype=dtype)

            if arg_count == 1 and isinstance(args[0], numbers.Number):
                obj.fill(args[0])
            else:
                i = 0
                for arg in args:
                    if i >= cls._shape[0]:
                        break
                    if isinstance(arg, (list, tuple, np.ndarray)):
                        arr = np.asarray(arg).flatten()
                        for e in arr:
                            if i >= cls._shape[0]:
                                break
                            else:
                                obj[i] = e
                                i += 1
                    elif isinstance(arg, numbers.Number):
                        obj[i] = arg
                        i += 1
                    else:
                        raise ValueError('Vector only accepts numeric values.')

            return obj.reshape(cls._shape).view(cls)
        else:
            arg_count = len(args)
            data = np.array([])
            for arg in args:
                if isinstance(arg, (list, tuple, np.ndarray)):
                    arr = np.asarray(arg).flatten()
                    data = np.concatenate([data, arr])
                elif isinstance(arg, numbers.Number):
                    data = np.append(data, arg)
                else:
                    raise ValueError('Vector only accepts numeric values.')
            cls._shape = (len(data), 1)
            return data.reshape(cls._shape).view(cls)

    def __array_finalize__(self, obj):
        if self._shape != obj.shape:
            obj.reshape(self._shape)

    def __add__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError(f"Cannot add a '{self.__class__.__name__}' with '{type(other).__name__}'")
        else:
            return np.add(self, other)

    def __iadd__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError(f"Cannot add a '{self.__class__.__name__}' with '{type(other).__name__}'")
        else:
            self[:] = np.add(self, other)
            return self

    def __sub__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError(f"Cannot subtract a '{self.__class__.__name__}' with '{type(other).__name__}'")
        else:
            return np.subtract(self, other)

    def __isub__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError(f"Cannot subtract a '{self.__class__.__name__}' with '{type(other).__name__}'")
        else:
            self[:] = np.subtract(self, other)
            return self

    def __div__(self, other):
        if not isinstance(other, numbers.Number):
            raise TypeError(f"Cannot divide a '{self.__class__.__name__}' by '{type(other).__name__}'")
        else:
            return np.divide(self, other)

    def __idiv__(self, other):
        if not isinstance(other, numbers.Number):
            raise TypeError(f"Cannot divide a '{self.__class__.__name__}' by '{type(other).__name__}'")
        else:
            self[:] = np.divide(self, other)
            return self

    def __mul__(self, other):
        if isinstance(other, numbers.Number):
            return np.multiply(self, other)
        elif isinstance(other, np.ndarray) and self.shape[1] == other.shape[0]:
            return np.matmul(self, other)
        else:
            raise TypeError(f"Cannot multiply a '{self.__class__.__name__}' with {type(other).__name__}")

    def __rmul__(self, other):
        if isinstance(other, numbers.Number):
            return np.multiply(other, self)
        elif isinstance(other, np.ndarray) and other.shape[1] == self.shape[0]:
            return np.matmul(other, self)
        else:
            raise TypeError(f"Cannot multiply a '{self.__class__.__name__}' with '{type(other).__name__}'")

    def __imul__(self, other):
        self[:] = self.__mul__(other)
        return self

    def __getitem__(self, i):
        return super(Vector, self).__getitem__((i, 0))

    def __setitem__(self, i, value):
        if 0 <= i < self._shape[0]:
            super(Vector, self).__setitem__((i, 0), value)
        else:
            raise IndexError('Index out of range.')

    def __eq__(self, other):
        return np.all(np.isclose(np.asarray(self), np.asarray(other)))

    def __str__(self):
        _str = 'Vec{} [' + '{:>11.8f},' * (self._shape[0] - 1) + ' {:>11.8f}]'
        return _str.format(self._shape[0], *self.view(np.ndarray).flatten())

    def __repr__(self):
        return self.__str__()

    @property
    def dims(self):
        return self._shape[0]


class Vec2(Vector):
    _shape = (2, 1)

    def __new__(cls, *args, **kwargs):
        return super(Vec2, cls).__new__(cls, *args, **kwargs)

    @classmethod
    def from_vec3(cls, vec, dtype=None):
        """Create a Vector2 from a Vector3.
        """
        dtype = dtype if dtype else vec.dtype
        return cls(np.asarray(vec[:2]), dtype=dtype)

    @classmethod
    def from_vec4(cls, vec, dtype=None):
        """Create a Vector2 from a Vector4.
        """
        dtype = dtype if dtype else vec.dtype
        return cls(np.asarray(vec[:2]), dtype=dtype)

    @property
    def x(self):
        return super(Vector, self).__getitem__((0, 0))

    @x.setter
    def x(self, value):
        super(Vector, self).__setitem__((0, 0), value)

    @property
    def y(self):
        return super(Vector, self).__getitem__((1, 0))

    @y.setter
    def y(self, value):
        super(Vector, self).__setitem__((1, 0), value)

    @classmethod
    def unit_x(cls, dtype=float):
        return Vec2(1., 0., dtype=dtype)

    @classmethod
    def unit_y(cls, dtype=float):
        return Vec2(0., 1., dtype=dtype)


class Vec3(Vector):
    _shape = (3, 1)

    def __new__(cls, *args, **kwargs):
        return super(Vec3, cls).__new__(cls, *args, **kwargs)

    @classmethod
    def from_vec2(cls, vec, z=0, dtype=None):
        """Create a Vector3 from a Vector2.
        """
        dtype = dtype if dtype else vec.dtype
        return cls(np.append(np.asarray(vec), z), dtype=dtype)

    @classmethod
    def from_vec4(cls, vec, dtype=None):
        """Create a Vector3 from a Vector4.
        """
        dtype = dtype if dtype else vec.dtype
        return cls(np.asarray(vec[:3]), dtype=dtype)

    @property
    def x(self):
        return super(Vector, self).__getitem__((0, 0))

    @x.setter
    def x(self, value):
        super(Vector, self).__setitem__((0, 0), value)

    @property
    def y(self):
        return super(Vector, self).__getitem__((1, 0))

    @y.setter
    def y(self, value):
        super(Vector, self).__setitem__((1, 0), value)

    @property
    def z(self):
        return super(Vector, self).__getitem__((2, 0))

    @z.setter
    def z(self, value):
        super(Vector, self).__setitem__((2, 0), value)

    @classmethod
    def unit_x(cls, dtype=float):
        return Vec3(1., 0., 0., dtype=dtype)

    @classmethod
    def unit_y(cls, dtype=float):
        return Vec3(0., 1., 0., dtype=dtype)

    @classmethod
    def unit_z(cls, dtype=float):
        return Vec3(0., 0., 1., dtype=dtype)

    @property
    def norm_squared(self):
        return np.sum(self ** 2, axis=-1)

    @property
    def norm(self) -> float:
        return math.sqrt(self.dot(self))

    @property
    def length_squared(self):
        return np.sum(self ** 2, axis=-1)

    @property
    def length(self):
        return np.sqrt(np.sum(self ** 2))

    def dot(self, other):
        return self.x * other.x + self.y * other.y + self.z * other.z

    def cross(self, other):
        return cross(self, other)

    def normalise(self):
        self = self.normalised
        return self

    @property
    def normalised(self):
        n = self.norm
        if n == 0. or n == 1.:
            return Vec3(self)
        else:
            return Vec3(self.x / n, self.y / n, self.z / n)

    @property
    def length_recip(self):
        """Computes 1.0 / length(). For valid results, `self` must not be of length zero."""
        return 1. / self.length

    def distance(self, other):
        """Computes the Euclidean distance between two points in space."""
        return (self - other).length

    def project_onto(self, other):
        """Returns the vector projection of `self` onto `other`."""
        other_normalised = other.normalised
        return self.normalised.dot(other_normalised) * other_normalised * self.norm


class Vec4(Vector):
    _shape = (4, 1)

    def __new__(cls, *args, **kwargs):
        return super(Vec4, cls).__new__(cls, *args, **kwargs)

    @classmethod
    def from_vec2(cls, vec, z=0, w=0, dtype=None):
        """Create a Vector4 from a Vector2.

        Examples:
            >>> va = Vec2(0.2, 0.1)
            >>> vb = Vec4.from_vec2(va)
        """
        dtype = dtype if dtype else vec.dtype
        return cls(np.append(np.asarray(vec), [z, w]), dtype=dtype)

    @classmethod
    def from_vec3(cls, vec, w=0, dtype=None):
        """Create a Vector4 from a Vector2.
        """
        dtype = dtype if dtype else vec.dtype
        return cls(np.append(np.asarray(vec), w), dtype=dtype)

    @property
    def x(self):
        return super(Vector, self).__getitem__((0, 0))

    @x.setter
    def x(self, value):
        super(Vector, self).__setitem__((0, 0), value)

    @property
    def y(self):
        return super(Vector, self).__getitem__((1, 0))

    @y.setter
    def y(self, value):
        super(Vector, self).__setitem__((1, 0), value)

    @property
    def z(self):
        return super(Vector, self).__getitem__((2, 0))

    @z.setter
    def z(self, value):
        super(Vector, self).__setitem__((2, 0), value)

    @property
    def w(self):
        return super(Vector, self).__getitem__((3, 0))

    @w.setter
    def w(self, value):
        super(Vector, self).__setitem__((3, 0), value)

    @classmethod
    def unit_x(cls, dtype=float):
        return Vec4(1., 0., 0., 0., dtype=dtype)

    @classmethod
    def unit_y(cls, dtype=float):
        return Vec4(0., 1., 0., 0., dtype=dtype)

    @classmethod
    def unit_z(cls, dtype=float):
        return Vec4(0., 0., 1., 0., dtype=dtype)

    @classmethod
    def unit_w(cls, dtype=float):
        return Vec4(0., 0., 0., 1., dtype=dtype)


def dot(a: Vec3, b: Vec3):
    """Computes the dot product of two vectors.
    Args:
        a (Vec3): an Nd array with dimension 3.
        b (Vec3):  an Nd array with dimension 3.

    Returns:
        The dot product of two vectors.
    """
    return np.sum(a * b, axis=-1)


def cross(a: Vec3, b: Vec3):
    """Computes the cross product of two vectors.
    Args:
        a (Vec3): an Nd array with dimension 3.
        b (Vec3): an Nd array with dimension 3.

    Returns:
        The cross product of two vectors.
    """
    return Vec3(np.cross(a, b, axis=0).reshape(3))  # cross along the first axis


def interpolate(a: Vec3, b: Vec3, t: float):
    return a + (b - a) * t


import abc
from typing import Sequence

import numpy as np

from ..graphics.util import DrawingMode
from ..misc import Color, PaletteDefault


class Shape(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return hasattr(subclass, 'vertices') and \
               hasattr(subclass, 'vertex_count') and hasattr(subclass, 'indices') and \
               hasattr(subclass, 'index_count') and hasattr(subclass, 'drawing_mode')

    def __init__(self, vertex_count, colors: Sequence[Color]):
        self._is_dirty = False
        self._vertex_count = vertex_count

        colors_count = len(colors)
        if colors_count == 0:
            self._colors = list(PaletteDefault.GreenA.as_color())
        elif colors_count == 1:
            self._colors = list(colors)
        elif colors_count != vertex_count:
            raise ValueError(f'{self.__class__.__name__} has {vertex_count} vertices, '
                             f'but only {colors_count} points are specified with a color.')
        else:
            self._colors = list(colors[:vertex_count])

    @property
    def color(self):
        if len(self._colors) == 1:
            return self._colors[0]
        else:
            return self._colors

    @color.setter
    def color(self, new_color):
        if isinstance(new_color, (list, tuple)):
            if len(new_color) < 3:
                raise ValueError(
                    f'Triangle has three vertices, but only {len(new_color)} points are specified with new color.')
            self._colors = new_color[:3]
            self._is_dirty = True
        elif isinstance(new_color, Color):
            self._colors = [new_color]
            self._is_dirty = True

    @property
    def colors(self):
        if len(self._colors) == 1:
            return np.tile(self._colors[0].rgba, self._vertex_count).ravel()
        else:
            colors = np.array([], dtype=np.float32)
            for color in self._colors:
                colors = np.concatenate([colors, color.rgba.astype(np.float32)])
            return colors.ravel()

    @property
    @abc.abstractmethod
    def vertices(self) -> np.ndarray:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def vertex_count(self) -> int:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def indices(self):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def index_count(self):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def drawing_mode(self) -> DrawingMode:
        raise NotImplementedError

    @property
    def interleaved_vertices(self) -> np.ndarray:
        """
        Returns interleaved data(position and color) for vertices.
        :return:
        """
        vertices_data = np.zeros(7 * self.vertex_count, dtype=np.float32)

        for i in range(0, self.vertex_count):
            index = i * 7
            vertices_data.put(list(range(index, index + 3)), self.vertices[i * 3: i * 3 + 3])
            vertices_data.put(list(range(index + 3, index + 7)), self.colors[i * 4: i * 4 + 4])

        return vertices_data

    @property
    def is_dirty(self):
        return self._is_dirty

    @is_dirty.setter
    def is_dirty(self, value):
        self._is_dirty = value




import logging
import os

from .window import Window

__all__ = [
    'current_time',
    'current_window',
    'gl_context_version',
    'init',
    'run',
]

logging.basicConfig(level=os.environ.get('LOGLEVEL', 'ERROR'))

__window__ = None


def gl_context_version():
    return None if __window__ is None else __window__.current_context_version()


def current_window():
    if __window__ is not None:
        return __window__
    else:
        raise ValueError('Window has not been created.')


def current_time():
    return current_window().elapsed_time


def init(win: Window):
    global __window__
    """Initialize the main loop."""
    if not isinstance(win, window.Window):
        raise ValueError('Not a valid window object.')
    __window__ = win
    __window__.activate()
    __window__.dispatch('on_init')
    __window__.dispatch('on_resize', win.width, win.height)


def run():
    """Run the main loop."""
    if __window__:
        __window__.main_loop()



from typing import Sequence

import numpy as np

from .shape import Shape
from ..graphics.util import DrawingMode
from ..math import Vec3
from ..misc import Color, PaletteDefault


class Ray(Shape):
    def __init__(self, origin, direction, colors: Sequence[Color] = (PaletteDefault.GreenB.as_color(),)):
        super().__init__(2, colors)
        self._o = Vec3(origin) if not isinstance(origin, Vec3) else origin
        self._d = Vec3(direction) if not isinstance(direction, Vec3) else direction

    def __str__(self):
        _str = 'Ray ⚬ {} ⟶ {}'
        return _str.format(self._o, self._d)

    @property
    def vertices(self) -> np.ndarray:
        return np.concatenate([self._o, self._o + self._d]).ravel()

    @property
    def vertex_count(self) -> int:
        return 2

    @property
    def drawing_mode(self) -> DrawingMode:
        return DrawingMode.Lines

    @property
    def origin(self):
        return self._o

    @origin.setter
    def origin(self, value: Vec3):
        self._o = value
        self._is_dirty = True

    @property
    def direction(self):
        return self._d

    @direction.setter
    def direction(self, value: Vec3):
        self._d = value
        self._is_dirty = True

    @property
    def indices(self):
        return np.array([0, 1], dtype=int)

    @property
    def index_count(self):
        return 2




__all__ = [
    'BufferBindingTarget',
    'BufferStorageFlag',
    'AccessPolicy',
    'DataUsage',
    'VertexBuffer',
    'IndexBuffer',
    'Buffer',
    'ShaderProgram',
    'ShaderType',
    'Shader',
    'VertexShader',
    'FragmentShader',
    'PixelShader',
    'DrawingMode',
    'VertexArrayObject',
    'VertexLayout',
    'VertexAttrib',
    'VertexAttribFormat',
    'VertexAttribDescriptor',
    'AreaLight',
    'DirectionalLight',
    'SpotLight',
    'PointLight',
    'draw'
]

from . import legacy
from .modern import draw
from .vertex_layout import *
from .buffer import *
from .array import *
from .program import *
from .shader import *
from .util import *
from .lights import *




import ctypes
import sys

import glfw
import numpy as np

from .util import DataUsage, BufferBindingTarget, GpuObject, BindSemanticObject
from .vertex_layout import VertexLayout
from .. import gl


class Buffer(GpuObject, BindSemanticObject):
    """Representation of an OpenGL buffer object.

    Doesn't hold a memory from cpu side.
    """
    def __init__(self, size: int, target: BufferBindingTarget, usage: DataUsage, mutable=True):
        super().__init__(target.value, -1)
        self._size = size

        if target not in BufferBindingTarget:
            raise ValueError(f'{target} is not a valid OpenGL buffer binding target.')

        if usage not in DataUsage:
            raise ValueError(f'{usage} is not a valid usage.')

        self._target = target.value
        self._usage = usage.value
        self._mutable = mutable

        # Creates a OpenGL buffer object.
        self._id = gl.glGenBuffers(1)

        # Set up its internal state.
        gl.glBindBuffer(self._target, self._id)

        if (not self._mutable) and (gl.current_context_version()[0] >= 4 and gl.current_context_version()[1] >= 4):
            gl.glBufferStorage(self._target, self._size, None, flags=None)
        else:
            gl.glBufferData(self._target, self._size, None, self._usage)

    def _delete(self):
        if self.is_valid():
            gl.glDeleteBuffers(1, [self._id])

    def _activate(self):
        self.bind()

    def _deactivate(self):
        self.unbind()

    @property
    def target(self) -> gl.Constant:
        return self._target

    @property
    def size(self):
        """Returns the size in bytes of the buffer."""
        return self._size

    # TODO: layout related

    def bind(self):
        """Bind this buffer to its OpenGL target."""
        gl.glBindBuffer(self._target, self._id)

    def unbind(self):
        """Reset the buffer's OpenGL target."""
        gl.glBindBuffer(self._target, 0)

    def set_data(self, data: np.ndarray, offset: int = 0, size: int = -1):
        """Update the contents of the buffer.

        Note:
            No boundary check for input data.

        Args:
            data (array like):
                Specifies a pointer to the new data that will be copied into the buffer..

            offset (int):
                Specifies the offset of the buffer where data replacement will begin, measured in bytes.

            size (int):
                Specifies the size in bytes of the data store region being replaced.
                Defaults to -1, the actual size will be the size of buffer.

        Returns:
            None
        """
        if not self._mutable:
            print('Warning: trying to update immutable buffer.', file=sys.stderr)

        gl.glBindBuffer(self._target, self._id)

        if offset == 0 and size == -1:
            gl.glBufferData(self._target, self._size, data, self._usage)
        else:
            gl.glBufferSubData(self._target, offset, size, data)

        gl.glBindBuffer(self._target, 0)

    def map(self, offset=0, size=0, ptr_type=ctypes.POINTER(ctypes.c_byte)):
        """Maps the entire buffer into system memory.

        TODO: test

        Args:
            offset:
                Byte offset from the beginning of the buffer memory.

            size:
                The size of the memory range (whole range by default) to map.

            ptr_type:
                Desired
        Returns:
            Pointer to the mapped range.
        """
        gl.glBindBuffer(self._target, self._id)

        # map whole buffer
        if offset == 0 and size == 0:
            ptr = ctypes.cast(gl.glMapBuffer(self._target, gl.GL_WRITE_ONLY),
                              ctypes.POINTER(ptr_type * self._size)).contents
            return ptr
        else:
            if offset > self._size or offset + size > self._size:
                print('Warning: trying to map the buffer memory with a range greater than its size.', file=sys.stderr)
            ptr = ctypes.cast(gl.glMapBufferRange(self._target, offset, size, gl.GL_MAP_WRITE_BIT), ptr_type).contents
            gl.glUnmapBuffer(self._target)
            return ptr

    def unmap(self):
        """Unmaps mapped memory."""
        gl.glUnmapBuffer(self._target)

    def resize(self, new_size):
        # map -> copy -> init
        dst = (ctypes.c_byte * new_size)()
        gl.glBindBuffer(self._target, self._id)
        data_ptr = gl.glMapBuffer(self._target, gl.GL_READ_ONLY)
        ctypes.memmove(dst, data_ptr, min(new_size, self._size))
        gl.glUnmapBuffer(self._target)

        self._size = new_size
        gl.glBufferData(self._target, self._size, dst, self._usage)


class VertexBuffer(Buffer):
    """Buffer for vertex attribute data."""
    def __init__(self, count, layout: VertexLayout, usage=DataUsage.StaticDraw):
        """Creates a Vertex Buffer Object.

        Args:
            count (int): Number of vertices.
            layout (VertexLayout): Attributes and its format attached to vertices.
            usage (DataUsage): Data usage of vertex buffer.
        """
        super(VertexBuffer, self).__init__(layout.compute_buffer_size(count), BufferBindingTarget.ArrayBuffer, usage)
        self._layout = layout

    @property
    def layout(self):
        return self._layout

    @layout.setter
    def layout(self, new_layout: VertexLayout):
        self._layout = new_layout

    @staticmethod
    def empty():
        return IndexBuffer(0, DataUsage.DynamicDraw)


class IndexBuffer(Buffer):
    """Buffer for index data."""
    def __init__(self, count, usage=DataUsage.StaticDraw):
        """Creates a Buffer object holding indices for drawing. By default, use `np.uint32` as index data type.

        Args:
            count (int): Number of indices.
            usage (DataUsage): Data usage of index buffer.
        """
        self._index_count = count
        size = count * np.dtype(np.uint32).itemsize
        super(IndexBuffer, self).__init__(size, BufferBindingTarget.ElementArrayBuffer, usage)

    @property
    def index_count(self):
        return self._index_count
    
    def set_data(self, data: np.ndarray, offset: int = 0, size: int = -1):
        super(IndexBuffer, self).set_data(data, offset, size)
        self._index_count = len(data.ravel())

    @staticmethod
    def empty():
        return IndexBuffer(0, DataUsage.DynamicDraw)



import enum
import os.path
import re
import sys

from .util import ShaderCodeParser, GpuObject
from .. import gl
from ..assets.resolver import PathResolver


class ShaderType(enum.Enum):
    Vertex = gl.GL_VERTEX_SHADER
    Fragment = gl.GL_FRAGMENT_SHADER
    Pixel = gl.GL_FRAGMENT_SHADER


class Shader(GpuObject):
    def __init__(self, shader_type: ShaderType, code: str, origin='<string>', is_file=False) -> None:
        """Creates an OpenGL shbader.

        Args:
            shader_type (ShaderType):
                The shader type.

            code (str):
                The source code string or the filepath to shader source code.
        """
        super().__init__(shader_type.value, -1)

        if is_file:
            if not (os.path.isfile(code) and os.path.exists(code)):
                msg = f'Shader source file {code} does not exist or is not a valid file path.'
                raise ValueError(msg)
            with open(code, 'rt') as file:
                self._code, self._uniforms, self._attribs = ShaderCodeParser.preprocess(file.read())
                self._origin = origin
        else:
            self._code, self._uniforms, self._attribs = ShaderCodeParser.preprocess(code)
            self._origin = origin

        self._create()
        self._compile()

    @staticmethod
    def _from_file(shader_type, filepath, resolver=PathResolver()):
        filepath = resolver.resolve(filepath) if not os.path.isabs(filepath) else filepath
        if not (os.path.isfile(filepath) and os.path.exists(filepath)):
            msg = f'Shader source file {os.path.abspath(filepath)} does not exist or is not a valid file path.'
            raise ValueError(msg)

        with open(filepath, 'rt') as file:
            return Shader(shader_type,
                          code=file.read(),
                          origin=os.path.basename(filepath),
                          is_file=False)

    @property
    def code(self) -> str:
        """
        Shader source code.

        Returns:
            GLSL code string.
        """
        return self._code

    @property
    def uniforms(self):
        """
        Shader uniforms obtained from source code.

        Returns:
            A list of names of uniforms.
        """
        return self._uniforms

    @property
    def attributes(self):
        """Shader input vertex attributes.

        Returns:
            A list of names of vertex attributes.
        """
        return self._attribs

    @property
    def handle(self):
        return self._id

    def _create(self):
        if not self._code:
            raise RuntimeError("No code has been given during shader creation.")

        if self._id != -1:
            print("Trying to recreate a existing shader: SKIP.", file=sys.stderr)

        self._id = gl.glCreateShader(self._type)

        if self._id <= 0:
            raise RuntimeError("Cannot create shader object.")

    def _compile(self):
        gl.glShaderSource(self._id, self._code)
        gl.glCompileShader(self._id)
        status = gl.glGetShaderiv(self._id, gl.GL_COMPILE_STATUS)
        if not status:
            error = str(gl.glGetShaderInfoLog(self._id))
            print('\033[93m' + error + '\033[0m', file=sys.stderr)
            raise RuntimeError("Shader compilation error.")

    def _delete(self):
        gl.glDeleteShader(self._id)


class VertexShader(Shader):
    def __init__(self, code: str):
        super(VertexShader, self).__init__(ShaderType.Vertex, code, '<string>', False)

    @classmethod
    def from_file(cls, filepath):
        return super(VertexShader, cls)._from_file(ShaderType.Vertex, filepath)

    def __repr__(self):
        return f"VertexShader ({self._origin})"


class FragmentShader(Shader):
    def __init__(self, code: str):
        super(FragmentShader, self).__init__(ShaderType.Fragment, code, '<string>', False)

    @classmethod
    def from_file(cls, filepath):
        return super(FragmentShader, cls)._from_file(ShaderType.Fragment, filepath)

    def __repr__(self):
        return f"FragmentShader ({self._origin})"


class PixelShader(Shader):
    def __init__(self, code: str):
        super(PixelShader, self).__init__(ShaderType.Fragment, code, '<string>', False)

    @classmethod
    def from_file(cls, filepath):
        return super(PixelShader, cls)._from_file(ShaderType.Fragment, filepath)

    def __repr__(self):
        return f"PixelShader ({self._origin})"



from enum import IntEnum, IntFlag

import glfw


class MouseButton(IntFlag):
    NONE = 0
    Left = 1 << 1
    Middle = 1 << 2
    Right = 1 << 3

    @classmethod
    def from_glfw_mouse_btn_code(cls, btn_code):
        return _MOUSE_MAP[btn_code]


class KeyModifier(IntFlag):
    NONE = 0
    Shift = 1 << 0
    Ctrl = 1 << 1
    Alt = 1 << 2
    Super = 1 << 5

    @classmethod
    def from_glfw_modifiers(cls, mods):
        modifiers = cls.NONE
        if mods & glfw.MOD_SHIFT:
            modifiers |= cls.Shift
        if mods & glfw.MOD_CONTROL:
            modifiers |= cls.Ctrl
        if mods & glfw.MOD_ALT:
            modifiers |= cls.Alt
        if mods & glfw.MOD_SUPER:
            modifiers |= cls.Super
        return modifiers


class KeyCode(IntEnum):
    Space = 32
    Apostrophe = 39  # '
    Comma = 44  # ,
    Minus = 45  # -
    Period = 46  # .
    Slash = 47  # /
    Num0 = 48
    Num1 = 49
    Num2 = 50
    Num3 = 51
    Num4 = 52
    Num5 = 53
    Num6 = 54
    Num7 = 55
    Num8 = 56
    Num9 = 57
    Semicolon = 59  # ;
    Equal = 61  # =
    A = 65
    B = 66
    C = 67
    D = 68
    E = 69
    F = 70
    G = 71
    H = 72
    I = 73
    J = 74
    K = 75
    L = 76
    M = 77
    N = 78
    O = 79
    P = 80
    Q = 81
    R = 82
    S = 83
    T = 84
    U = 85
    V = 86
    W = 87
    X = 88
    Y = 89
    Z = 90
    LeftBracket = 91  # [
    Backslash = 92  # \
    RightBracket = 93  # ]
    GraveAccent = 96  # `
    Escape = 256
    Enter = 257
    Tab = 258
    Backspace = 259
    Delete = 261
    Right = 262
    Left = 263
    Down = 264
    Up = 265
    PageUp = 266
    PageDown = 267
    Home = 268
    End = 269
    CapsLock = 280
    F1 = 290
    F2 = 291
    F3 = 292
    F4 = 293
    F5 = 294
    F6 = 295
    F7 = 296
    F8 = 297
    F9 = 298
    F10 = 299
    F11 = 300
    F12 = 301
    Keypad0 = 320
    Keypad1 = 321
    Keypad2 = 322
    Keypad3 = 323
    Keypad4 = 324
    Keypad5 = 325
    Keypad6 = 326
    Keypad7 = 327
    Keypad8 = 328
    Keypad9 = 329
    KeypadDecimal = 330
    KeypadDivide = 331
    KeypadMultiply = 332
    KeypadSubtract = 333
    KeypadAdd = 334
    KeypadEnter = 335
    KeypadEqual = 336
    LeftShift = 1 << 9
    LeftControl = 1 << 10
    LeftAlt = 1 << 11
    LeftSuper = 1 << 12
    RightShift = 1 << 13
    RightControl = 1 << 14
    RightAlt = 1 << 15
    RightSuper = 1 << 16
    Menu = 348

    @classmethod
    def from_glfw_keycode(cls, keycode):
        if keycode in _KEY_MAP:
            return _KEY_MAP[keycode]


_KEY_MAP = {
    glfw.KEY_SPACE: KeyCode.Space,
    glfw.KEY_APOSTROPHE: KeyCode.Apostrophe,
    glfw.KEY_COMMA: KeyCode.Comma,
    glfw.KEY_MINUS: KeyCode.Minus,
    glfw.KEY_PERIOD: KeyCode.Period,
    glfw.KEY_SLASH: KeyCode.Slash,
    glfw.KEY_0: KeyCode.Num0,
    glfw.KEY_1: KeyCode.Num1,
    glfw.KEY_2: KeyCode.Num2,
    glfw.KEY_3: KeyCode.Num3,
    glfw.KEY_4: KeyCode.Num4,
    glfw.KEY_5: KeyCode.Num5,
    glfw.KEY_6: KeyCode.Num6,
    glfw.KEY_7: KeyCode.Num7,
    glfw.KEY_8: KeyCode.Num8,
    glfw.KEY_9: KeyCode.Num9,
    glfw.KEY_SEMICOLON: KeyCode.Semicolon,
    glfw.KEY_EQUAL: KeyCode.Equal,
    glfw.KEY_A: KeyCode.A,
    glfw.KEY_B: KeyCode.B,
    glfw.KEY_C: KeyCode.C,
    glfw.KEY_D: KeyCode.D,
    glfw.KEY_E: KeyCode.E,
    glfw.KEY_F: KeyCode.F,
    glfw.KEY_G: KeyCode.G,
    glfw.KEY_H: KeyCode.H,
    glfw.KEY_I: KeyCode.I,
    glfw.KEY_J: KeyCode.J,
    glfw.KEY_K: KeyCode.K,
    glfw.KEY_L: KeyCode.L,
    glfw.KEY_M: KeyCode.M,
    glfw.KEY_N: KeyCode.N,
    glfw.KEY_O: KeyCode.O,
    glfw.KEY_P: KeyCode.P,
    glfw.KEY_Q: KeyCode.Q,
    glfw.KEY_R: KeyCode.R,
    glfw.KEY_S: KeyCode.S,
    glfw.KEY_T: KeyCode.T,
    glfw.KEY_U: KeyCode.U,
    glfw.KEY_V: KeyCode.V,
    glfw.KEY_W: KeyCode.W,
    glfw.KEY_X: KeyCode.X,
    glfw.KEY_Y: KeyCode.Y,
    glfw.KEY_Z: KeyCode.Z,
    glfw.KEY_LEFT_BRACKET: KeyCode.LeftBracket,
    glfw.KEY_BACKSLASH: KeyCode.Backslash,
    glfw.KEY_RIGHT_BRACKET: KeyCode.RightBracket,
    glfw.KEY_GRAVE_ACCENT: KeyCode.GraveAccent,
    glfw.KEY_ESCAPE: KeyCode.Escape,
    glfw.KEY_ENTER: KeyCode.Enter,
    glfw.KEY_TAB: KeyCode.Tab,
    glfw.KEY_BACKSPACE: KeyCode.Backspace,
    glfw.KEY_DELETE: KeyCode.Delete,
    glfw.KEY_RIGHT: KeyCode.Right,
    glfw.KEY_LEFT: KeyCode.Left,
    glfw.KEY_DOWN: KeyCode.Down,
    glfw.KEY_UP: KeyCode.Up,
    glfw.KEY_PAGE_UP: KeyCode.PageUp,
    glfw.KEY_PAGE_DOWN: KeyCode.PageDown,
    glfw.KEY_HOME: KeyCode.Home,
    glfw.KEY_END: KeyCode.End,
    glfw.KEY_CAPS_LOCK: KeyCode.CapsLock,
    glfw.KEY_F1: KeyCode.F1,
    glfw.KEY_F2: KeyCode.F2,
    glfw.KEY_F3: KeyCode.F3,
    glfw.KEY_F4: KeyCode.F4,
    glfw.KEY_F5: KeyCode.F5,
    glfw.KEY_F6: KeyCode.F6,
    glfw.KEY_F7: KeyCode.F7,
    glfw.KEY_F8: KeyCode.F8,
    glfw.KEY_F9: KeyCode.F9,
    glfw.KEY_F10: KeyCode.F10,
    glfw.KEY_F11: KeyCode.F11,
    glfw.KEY_F12: KeyCode.F12,
    glfw.KEY_KP_0: KeyCode.Keypad0,
    glfw.KEY_KP_1: KeyCode.Keypad1,
    glfw.KEY_KP_2: KeyCode.Keypad2,
    glfw.KEY_KP_3: KeyCode.Keypad3,
    glfw.KEY_KP_4: KeyCode.Keypad4,
    glfw.KEY_KP_5: KeyCode.Keypad5,
    glfw.KEY_KP_6: KeyCode.Keypad6,
    glfw.KEY_KP_7: KeyCode.Keypad7,
    glfw.KEY_KP_8: KeyCode.Keypad8,
    glfw.KEY_KP_9: KeyCode.Keypad9,
    glfw.KEY_KP_DECIMAL: KeyCode.KeypadDecimal,
    glfw.KEY_KP_DIVIDE: KeyCode.KeypadDivide,
    glfw.KEY_KP_MULTIPLY: KeyCode.KeypadMultiply,
    glfw.KEY_KP_SUBTRACT: KeyCode.KeypadSubtract,
    glfw.KEY_KP_ADD: KeyCode.KeypadAdd,
    glfw.KEY_KP_ENTER: KeyCode.KeypadEnter,
    glfw.KEY_KP_EQUAL: KeyCode.KeypadEqual,
    glfw.KEY_LEFT_SHIFT: KeyCode.LeftShift,
    glfw.KEY_LEFT_CONTROL: KeyCode.LeftControl,
    glfw.KEY_LEFT_ALT: KeyCode.LeftAlt,
    glfw.KEY_LEFT_SUPER: KeyCode.LeftSuper,
    glfw.KEY_RIGHT_SHIFT: KeyCode.RightShift,
    glfw.KEY_RIGHT_CONTROL: KeyCode.RightControl,
    glfw.KEY_RIGHT_ALT: KeyCode.RightAlt,
    glfw.KEY_RIGHT_SUPER: KeyCode.RightSuper,
    glfw.KEY_MENU: KeyCode.Menu
}

_MOUSE_MAP = {
    glfw.MOUSE_BUTTON_LEFT: MouseButton.Left,
    glfw.MOUSE_BUTTON_MIDDLE: MouseButton.Middle,
    glfw.MOUSE_BUTTON_RIGHT: MouseButton.Right
}



import numbers
from typing import Sequence

import numpy as np

from .shape import Shape
from ..graphics.util import DrawingMode
from ..math import Vec3
from ..misc import Color, PaletteDefault


class Line(Shape):
    def __init__(self, points: Sequence[Vec3], colors: Sequence[Color] = (PaletteDefault.BrownB.as_color(),)):
        super(Line, self).__init__(len(points), colors)
        self._points = np.array(points, dtype=np.float32).reshape((3, len(points)))
        self._points_count = len(points)

    def __getitem__(self, item: int):
        if not isinstance(item, numbers.Integral) or item < 0 or item > 2:
            raise IndexError(f'Invalid index {item}')
        return self._points[item].reshape((3, 1)).view(Vec3)

    def __iter__(self):
        self._i = 0
        return self

    def __next__(self):
        if self._i < 3:
            value = self._points[self._i]
            self._i += 1
            return value
        else:
            raise StopIteration

    @property
    def drawing_mode(self) -> DrawingMode:
        return DrawingMode.LineStrip

    @property
    def vertex_count(self) -> int:
        return self._points_count

    @property
    def vertices(self) -> np.ndarray:
        return np.array(self._points).ravel()

    @property
    def indices(self):
        return np.array(list(range(0, self._points_count)), dtype=int)

    @property
    def index_count(self):
        return self._points_count




import ctypes

from .util import GpuObject, BindSemanticObject
from .buffer import VertexBuffer, IndexBuffer
from .. import gl


# TODO: deal with ibo


class VertexArrayObject(GpuObject, BindSemanticObject):
    """
    Vertex Array Object manages client side states: VBO bindings, IBO bindings and vertex layout.

    For instance you can bind one VAO, then bind a VBO, an IBO and configure the vertex layout,
    bind another VAO, another VBO and IBO and configure a different vertex layout.

    Whenever you bind the first VAO, the VBO and IBO associated with it are bound and the associated
    vertex format is used. Binding the second VAO will then switch to the other pair of VBO/IBO and
    respective vertex layout.
    """
    def __init__(self):
        super().__init__('VertexBufferObject', -1)
        self._id = gl.glGenVertexArrays(1)

    def __enter__(self):
        self.bind()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.unbind()

    def _delete(self):
        gl.glDeleteVertexArrays(1, (self._id,))

    def bind_vertex_buffer(self, buffer: VertexBuffer, attrib_locations=None):
        """
        Args:
            buffer (VertexBuffer):
                Specifies the buffer used to define the array of vertex attribute data.

            attrib_locations (int or list[int]):
                Specifies the location of the generic vertex attribute to be enabled or disabled.
                This is helpful when you are using multiple buffers for different vertex attribute.
                If passed as a list, each attribute of the vertex will be bind to the corresponding
                location.

        Returns:
            None
        """
        self.bind()

        buffer.bind()

        locations = []

        if attrib_locations is None:
            locations = list(range(0, buffer.layout.attrib_count))
        elif isinstance(attrib_locations, int):
            locations = list(i + attrib_locations for i in range(0, buffer.layout.attrib_count))
        elif isinstance(attrib_locations, list):
            if len(attrib_locations) < buffer.layout.attrib_count:
                raise ValueError('Not sufficient attribute locations.')
            locations = attrib_locations[:buffer.layout.attrib_count]
        else:
            raise ValueError('Wrong attribute locations.')

        for i, (attrib, (fmt, dim)) in enumerate(buffer.layout.description.items()):
            gl.glEnableVertexAttribArray(locations[i])
            gl.glVertexAttribPointer(locations[i], dim, fmt.gl_type, gl.GL_FALSE,
                                     buffer.layout.stride,
                                     ctypes.c_void_p(buffer.layout.offset_of(attrib)))

        buffer.unbind()
        self.unbind()

    def bind_index_buffer(self, buffer: IndexBuffer):
        # TODO
        pass

    def _activate(self):
        self.bind()

    def _deactivate(self):
        self.unbind()

    def bind(self):
        gl.glBindVertexArray(self._id)

    def unbind(self):
        gl.glBindVertexArray(0)



import os

__all__ = [
    'default_resolver',
    'PathResolver',
    'Image',
    'default_asset_mgr',
    'AssetManager'
]

from .image import Image
from .resolver import PathResolver, default_resolver
from .manager import AssetManager, default_asset_mgr




import numbers
from enum import Enum
import numpy as np


def sign(n: numbers.Number):
    if n > 0:
        return 1
    else:
        return -1


def hex2rgb(color):
    _str = color.lstrip('#')
    return [float(int(_str[i:i + 2], 16)) / 255. for i in (0, 2, 4)]


def hex2rgba(color, alpha):
    _str = color.lstrip('#')
    return [float(int(_str[i:i + 2], 16)) / 255. for i in (0, 2, 4)] + [alpha]


class Color:
    """
    Represents a color using 4 normalised channels (red, green, blue and alpha).

    Examples:
        >>> c0 = Color()
        >>> c1 = Color('#aabbcc', alpha = 1.0)
        >>> c2 = Color((1.0, 0.2, 0.3), alpha=1.0)
        >>> c3 = Color([1.0, 0.2, 0.3], alpha=1.0)
    """
    def __init__(self, color=None, alpha=1.0):
        """
        Stores the values(between 0.0 and 1.0) of a RGBA color.

        Args:
            color (str, list, tuple or np.ndarray):
                Input color could be a hex string, a tuple or np.ndarray.
            alpha (float):
                Alpha channel of the color.
        """
        if isinstance(color, (list, tuple, np.ndarray)):
            if len(color) < 3:
                raise ValueError("Color needs at least 3 elements.")
            self._rgba = np.ones(4, dtype=np.float32) * (0., 0., 0., alpha)
            self._rgba[...] = color[...]
        elif isinstance(color, str):
            self._rgba = np.asarray(hex2rgba(color, alpha))
        else:
            self._rgba = np.zeros(4, dtype=np.float32)

    @property
    def rgba(self):
        return self._rgba

    @rgba.setter
    def rgba(self, rgba):
        self._rgba[:] = rgba[:4]

    @property
    def rgb(self):
        return self._rgba[:3]

    @rgb.setter
    def rgb(self, rgb):
        self._rgba[:3] = rgb[:3]

    def __getitem__(self, index):
        if index > 4 or index < 0:
            raise IndexError("Color has only 4 elements.")
        return self._rgba[index]

    def __setitem__(self, index, value):
        if index > 4 or index < 0:
            raise IndexError("Color has only 4 elements.")
        self._rgba[index] = np.clip(value, 0.0, 1.0)

    def __iter__(self):
        return self._rgba.__iter__()

    def __next__(self):
        self._index += 1
        if self._index >= len(self._rgba):
            self._index = -1
            raise StopIteration
        else:
            return self._rgba[self._index]


class PaletteMaterial(Enum):
    Red50 = '#ffebee'
    Red100 = '#ffcdd2'
    Red200 = '#ef9a9a'
    Red300 = '#e57373'
    Red400 = '#ef5350'
    Red500 = '#f44336'
    Red600 = '#e53935'
    Red700 = '#d32f2f'
    Red800 = '#c62828'
    Red900 = '#b71c1c'
    RedAccent100 = '#ff8a80'
    RedAccent200 = '#ff5252'
    RedAccent400 = '#ff1744'
    RedAccent700 = '#d50000'
    Pink50 = '#fce4ec'
    Pink100 = '#f8bbd0'
    Pink200 = '#f48fb1'
    Pink300 = '#f06292'
    Pink400 = '#ec407a'
    Pink500 = '#e91e63'
    Pink600 = '#d81b60'
    Pink700 = '#c2185b'
    Pink800 = '#ad1457'
    Pink900 = '#880e4f'
    PinkAccent100 = '#ff80ab'
    PinkAccent200 = '#ff4081'
    PinkAccent400 = '#f50057'
    PinkAccent700 = '#c51162'
    Purple50 = '#f3e5f5'
    Purple100 = '#e1bee7'
    Purple200 = '#ce93d8'
    Purple300 = '#ba68c8'
    Purple400 = '#ab47bc'
    Purple500 = '#9c27b0'
    Purple600 = '#8e24aa'
    Purple700 = '#7b1fa2'
    Purple800 = '#6a1b9a'
    Purple900 = '#4a148c'
    PurpleAccent100 = '#ea80fc'
    PurpleAccent200 = '#e040fb'
    PurpleAccent400 = '#d500f9'
    PurpleAccent700 = '#aa00ff'
    DeepPurple50 = '#ede7f6'
    DeepPurple100 = '#d1c4e9'
    DeepPurple200 = '#b39ddb'
    DeepPurple300 = '#9575cd'
    DeepPurple400 = '#7e57c2'
    DeepPurple500 = '#673ab7'
    DeepPurple600 = '#5e35b1'
    DeepPurple700 = '#512da8'
    DeepPurple800 = '#4527a0'
    DeepPurple900 = '#311b92'
    DeepPurpleAccent100 = '#b388ff'
    DeepPurpleAccent200 = '#7c4dff'
    DeepPurpleAccent400 = '#651fff'
    DeepPurpleAccent700 = '#6200ea'
    Indigo50 = '#e8eaf6'
    Indigo100 = '#c5cae9'
    Indigo200 = '#9fa8da'
    Indigo300 = '#7986cb'
    Indigo400 = '#5c6bc0'
    Indigo500 = '#3f51b5'
    Indigo600 = '#3949ab'
    Indigo700 = '#303f9f'
    Indigo800 = '#283593'
    Indigo900 = '#1a237e'
    IndigoAccent100 = '#8c9eff'
    IndigoAccent200 = '#536dfe'
    IndigoAccent400 = '#3d5afe'
    IndigoAccent700 = '#304ffe'
    Blue50 = '#e3f2fd'
    Blue100 = '#bbdefb'
    Blue200 = '#90caf9'
    Blue300 = '#64b5f6'
    Blue400 = '#42a5f5'
    Blue500 = '#2196f3'
    Blue600 = '#1e88e5'
    Blue700 = '#1976d2'
    Blue800 = '#1565c0'
    Blue900 = '#0d47a1'
    BlueAccent100 = '#82b1ff'
    BlueAccent200 = '#448aff'
    BlueAccent400 = '#2979ff'
    BlueAccent700 = '#2962ff'
    LightBlue50 = '#e1f5fe'
    LightBlue100 = '#b3e5fc'
    LightBlue200 = '#81d4fa'
    LightBlue300 = '#4fc3f7'
    LightBlue400 = '#29b6f6'
    LightBlue500 = '#03a9f4'
    LightBlue600 = '#039be5'
    LightBlue700 = '#0288d1'
    LightBlue800 = '#0277bd'
    LightBlue900 = '#01579b'
    LightBlueAccent100 = '#80d8ff'
    LightBlueAccent200 = '#40c4ff'
    LightBlueAccent400 = '#00b0ff'
    LightBlueAccent700 = '#0091ea'
    Cyan50 = '#e0f7fa'
    Cyan100 = '#b2ebf2'
    Cyan200 = '#80deea'
    Cyan300 = '#4dd0e1'
    Cyan400 = '#26c6da'
    Cyan500 = '#00bcd4'
    Cyan600 = '#00acc1'
    Cyan700 = '#0097a7'
    Cyan800 = '#00838f'
    Cyan900 = '#006064'
    CyanAccent100 = '#84ffff'
    CyanAccent200 = '#18ffff'
    CyanAccent400 = '#00e5ff'
    CyanAccent700 = '#00b8d4'
    Teal50 = '#e0f2f1'
    Teal100 = '#b2dfdb'
    Teal200 = '#80cbc4'
    Teal300 = '#4db6ac'
    Teal400 = '#26a69a'
    Teal500 = '#009688'
    Teal600 = '#00897b'
    Teal700 = '#00796b'
    Teal800 = '#00695c'
    Teal900 = '#004d40'
    TealAccent100 = '#a7ffeb'
    TealAccent200 = '#64ffda'
    TealAccent400 = '#1de9b6'
    TealAccent700 = '#00bfa5'
    Green50 = '#e8f5e9'
    Green100 = '#c8e6c9'
    Green200 = '#a5d6a7'
    Green300 = '#81c784'
    Green400 = '#66bb6a'
    Green500 = '#4caf50'
    Green600 = '#43a047'
    Green700 = '#388e3c'
    Green800 = '#2e7d32'
    Green900 = '#1b5e20'
    GreenAccent100 = '#b9f6ca'
    GreenAccent200 = '#69f0ae'
    GreenAccent400 = '#00e676'
    GreenAccent700 = '#00c853'
    LightGreen50 = '#f1f8e9'
    LightGreen100 = '#dcedc8'
    LightGreen200 = '#c5e1a5'
    LightGreen300 = '#aed581'
    LightGreen400 = '#9ccc65'
    LightGreen500 = '#8bc34a'
    LightGreen600 = '#7cb342'
    LightGreen700 = '#689f38'
    LightGreen800 = '#558b2f'
    LightGreen900 = '#33691e'
    LightGreenAccent100 = '#ccff90'
    LightGreenAccent200 = '#b2ff59'
    LightGreenAccent400 = '#76ff03'
    LightGreenAccent700 = '#64dd17'
    Lime50 = '#f9fbe7'
    Lime100 = '#f0f4c3'
    Lime200 = '#e6ee9c'
    Lime300 = '#dce775'
    Lime400 = '#d4e157'
    Lime500 = '#cddc39'
    Lime600 = '#c0ca33'
    Lime700 = '#afb42b'
    Lime800 = '#9e9d24'
    Lime900 = '#827717'
    LimeAccent100 = '#f4ff81'
    LimeAccent200 = '#eeff41'
    LimeAccent400 = '#c6ff00'
    LimeAccent700 = '#aeea00'
    Yellow50 = '#fffde7'
    Yellow100 = '#fff9c4'
    Yellow200 = '#fff59d'
    Yellow300 = '#fff176'
    Yellow400 = '#ffee58'
    Yellow500 = '#ffeb3b'
    Yellow600 = '#fdd835'
    Yellow700 = '#fbc02d'
    Yellow800 = '#f9a825'
    Yellow900 = '#f57f17'
    YellowAccent100 = '#ffff8d'
    YellowAccent200 = '#ffff00'
    YellowAccent400 = '#ffea00'
    YellowAccent700 = '#ffd600'
    Amber50 = '#fff8e1'
    Amber100 = '#ffecb3'
    Amber200 = '#ffe082'
    Amber300 = '#ffd54f'
    Amber400 = '#ffca28'
    Amber500 = '#ffc107'
    Amber600 = '#ffb300'
    Amber700 = '#ffa000'
    Amber800 = '#ff8f00'
    Amber900 = '#ff6f00'
    AmberAccent100 = '#ffe57f'
    AmberAccent200 = '#ffd740'
    AmberAccent400 = '#ffc400'
    AmberAccent700 = '#ffab00'
    Orange50 = '#fff3e0'
    Orange100 = '#ffe0b2'
    Orange200 = '#ffcc80'
    Orange300 = '#ffb74d'
    Orange400 = '#ffa726'
    Orange500 = '#ff9800'
    Orange600 = '#fb8c00'
    Orange700 = '#f57c00'
    Orange800 = '#ef6c00'
    Orange900 = '#e65100'
    OrangeAccent100 = '#ffd180'
    OrangeAccent200 = '#ffab40'
    OrangeAccent400 = '#ff9100'
    OrangeAccent700 = '#ff6d00'
    DeepOrange50 = '#fbe9e7'
    DeepOrange100 = '#ffccbc'
    DeepOrange200 = '#ffab91'
    DeepOrange300 = '#ff8a65'
    DeepOrange400 = '#ff7043'
    DeepOrange500 = '#ff5722'
    DeepOrange600 = '#f4511e'
    DeepOrange700 = '#e64a19'
    DeepOrange800 = '#d84315'
    DeepOrange900 = '#bf360c'
    DeepOrangeAccent100 = '#ff9e80'
    DeepOrangeAccent200 = '#ff6e40'
    DeepOrangeAccent400 = '#ff3d00'
    DeepOrangeAccent700 = '#dd2c00'
    Brown50 = '#efebe9'
    Brown100 = '#d7ccc8'
    Brown200 = '#bcaaa4'
    Brown300 = '#a1887f'
    Brown400 = '#8d6e63'
    Brown500 = '#795548'
    Brown600 = '#6d4c41'
    Brown700 = '#5d4037'
    Brown800 = '#4e342e'
    Brown900 = '#3e2723'
    Grey50 = '#fafafa'
    Grey100 = '#f5f5f5'
    Grey200 = '#eeeeee'
    Grey300 = '#e0e0e0'
    Grey400 = '#bdbdbd'
    Grey500 = '#9e9e9e'
    Grey600 = '#757575'
    Grey700 = '#616161'
    Grey800 = '#424242'
    Grey900 = '#212121'
    BlueGrey50 = '#eceff1'
    BlueGrey100 = '#cfd8dc'
    BlueGrey200 = '#b0bec5'
    BlueGrey300 = '#90a4ae'
    BlueGrey400 = '#78909c'
    BlueGrey500 = '#607d8b'
    BlueGrey600 = '#546e7a'
    BlueGrey700 = '#455a64'
    BlueGrey800 = '#37474f'
    BlueGrey900 = '#263238'
    Black = '#000000'
    White = '#ffffff'

    def as_color(self, alpha=1.0):
        return Color(self.value, alpha)

    def __iter__(self):
        return Color(self.value, 1.0).__iter__()


class PaletteWeb(Enum):
    Black = '#000000'
    Silver = '#c0c0c0'
    Gray = '#808080'
    White = '#ffffff'
    Maroon = '#800000'
    Red = '#ff0000'
    Purple = '#800080'
    Fuchsia = '#ff00ff'
    Green = '#008000'
    Lime = '#00ff00'
    Olive = '#808000'
    Yellow = '#ffff00'
    Navy = '#000080'
    Blue = '#0000ff'
    Teal = '#008080'
    Aqua = '#00ffff'

    def as_color(self, alpha=1.0):
        return Color(self.value, alpha)

    def __iter__(self):
        return Color(self.value, 1.0).__iter__()


class PaletteSvg(Enum):
    AliceBlue = '#f0f8ff'
    AntiqueWhite = '#faebd7'
    Aqua = '#00ffff'
    Aquamarine = '#7fffd4'
    Azure = '#f0ffff'
    Beige = '#f5f5dc'
    Bisque = '#ffe4c4'
    Black = '#000000'
    BlanchedAlmond = '#ffebcd'
    Blue = '#0000ff'
    BlueViolet = '#8a2be2'
    Brown = '#a52a2a'
    Burlywood = '#deb887'
    CadetBlue = '#5f9ea0'
    Chartreuse = '#7fff00'
    Chocolate = '#d2691e'
    Coral = '#ff7f50'
    CornflowerBlue = '#6495ed'
    CornSilk = '#fff8dc'
    Crimson = '#dc143c'
    Cyan = '#00ffff'
    DarkBlue = '#00008b'
    DarkCyan = '#008b8b'
    DarkGoldenRod = '#b8860b'
    DarkGray = '#a9a9a9'
    DarkGrey = '#a9a9a9'
    DarkGreen = '#006400'
    DarkKhaki = '#bdb76b'
    DarkMagenta = '#8b008b'
    DarkOliveGreen = '#556b2f'
    DarkOrange = '#ff8c00'
    DarkOrchid = '#9932cc'
    DarkRed = '#8b0000'
    DarkSalmon = '#e9967a'
    DarkSeaGreen = '#8fbc8f'
    DarkSlateBlue = '#483d8b'
    DarkSlateGray = '#2f4f4f'
    DarkSlateGrey = '#2f4f4f'
    DarkTurquoise = '#00ced1'
    DarkViolet = '#9400d3'
    DeepPink = '#ff1493'
    DeepSkyBlue = '#00bfff'
    DimGray = '#696969'
    DimGrey = '#696969'
    DodgerBlue = '#1e90ff'
    Firebrick = '#b22222'
    FloralWhite = '#fffaf0'
    ForestGreen = '#228b22'
    Fuchsia = '#ff00ff'
    Gainsboro = '#dcdcdc'
    GhostWhite = '#f8f8ff'
    Gold = '#ffd700'
    GoldenRod = '#daa520'
    Gray = '#808080'
    Grey = '#808080'
    Green = '#008000'
    GreenYellow = '#adff2f'
    Honeydew = '#f0fff0'
    HotPink = '#ff69b4'
    IndianRed = '#cd5c5c'
    Indigo = '#4b0082'
    Ivory = '#fffff0'
    Khaki = '#f0e68c'
    Lavender = '#e6e6fa'
    LavenderBlush = '#fff0f5'
    LawnGreen = '#7cfc00'
    LemonChiffon = '#fffacd'
    LightBlue = '#add8e6'
    LightCoral = '#f08080'
    LightCyan = '#e0ffff'
    LightGoldenRodYellow = '#fafad2'
    LightGray = '#d3d3d3'
    LightGrey = '#d3d3d3'
    LightGreen = '#90ee90'
    LightPink = '#ffb6c1'
    LightSalmon = '#ffa07a'
    LightSeaGreen = '#20b2aa'
    LightSkyBlue = '#87cefa'
    LightSlateGray = '#778899'
    LightSlateGrey = '#778899'
    LightSteelBlue = '#b0c4de'
    LightYellow = '#ffffe0'
    Lime = '#00ff00'
    LimeGreen = '#32cd32'
    Linen = '#faf0e6'
    Magenta = '#ff00ff'
    Maroon = '#800000'
    MediumAquamarine = '#66cdaa'
    MediumBlue = '#0000cd'
    MediumOrchid = '#ba55d3'
    MediumPurple = '#9370d8'
    MediumSeaGreen = '#3cb371'
    MediumSlateBlue = '#7b68ee'
    MediumSpringGreen = '#00fa9a'
    MediumTurquoise = '#48d1cc'
    MediumVioletRed = '#c71585'
    MidnightBlue = '#191970'
    MintCream = '#f5fffa'
    MistyRose = '#ffe4e1'
    Moccasin = '#ffe4b5'
    NavajoWhite = '#ffdead'
    Navy = '#000080'
    Oldlace = '#fdf5e6'
    Olive = '#808000'
    OlivedRab = '#6b8e23'
    Orange = '#ffa500'
    OrangeRed = '#ff4500'
    Orchid = '#da70d6'
    PaleGoldenRod = '#eee8aa'
    PaleGreen = '#98fb98'
    PaleTurquoise = '#afeeee'
    PaleVioletRed = '#d87093'
    Papayawhip = '#ffefd5'
    Peachpuff = '#ffdab9'
    Peru = '#cd853f'
    Pink = '#ffc0cb'
    Plum = '#dda0dd'
    PowderBlue = '#b0e0e6'
    Purple = '#800080'
    Red = '#ff0000'
    RosyBrown = '#bc8f8f'
    RoyalBlue = '#4169e1'
    SaddleBrown = '#8b4513'
    Salmon = '#fa8072'
    SandyBrown = '#f4a460'
    SeaGreen = '#2e8b57'
    Seashell = '#fff5ee'
    Sienna = '#a0522d'
    Silver = '#c0c0c0'
    SkyBlue = '#87ceeb'
    SlateBlue = '#6a5acd'
    SlateGray = '#708090'
    SlateGrey = '#708090'
    Snow = '#fffafa'
    SpringGreen = '#00ff7f'
    SteelBlue = '#4682b4'
    Tan = '#d2b48c'
    Teal = '#008080'
    Thistle = '#d8bfd8'
    Tomato = '#ff6347'
    Turquoise = '#40e0d0'
    Violet = '#ee82ee'
    Wheat = '#f5deb3'
    White = '#ffffff'
    WhiteSmoke = '#f5f5f5'
    Yellow = '#ffff00'
    YellowGreen = '#9acd32'

    def as_color(self, alpha=1.0):
        return Color(self.value, alpha)

    def __iter__(self):
        return Color(self.value, 1.0).__iter__()


class PaletteDefault(Enum):
    Background = '#1c1c1c'
    BlackA = '#3d352a'
    BlackB = '#554444'
    RedA = '#cd5c5c'
    RedB = '#cc5533'
    GreenA = '#86af80'
    GreenB = '#88aa22'
    YellowA = '#e8ae5b'
    YellowB = '#ffa75d'
    BlueA = '#6495ed'
    BlueB = '#87ceeb'
    BrownA = '#996600'
    BrownB = '#deb887'
    GreyA = '#b0c4de'
    GreyB = '#d0d0d0'
    WhiteA = '#bbaa99'
    WhiteB = '#ddccbb'

    def as_color(self, alpha=1.0):
        return Color(self.value, alpha)

    def __iter__(self):
        return Color(self.value, 1.0).__iter__()



import os.path
from dataclasses import dataclass

import numpy as np

from .texture import Texture


# TODO: implement texture properties


@dataclass(frozen=False)
class MaterialData:
    __slots__ = ('name', 'ambient', 'diffuse', 'specular', 'diffuse_map', 'bump_map', 'normal_map',
                 'shininess', 'refractive_index', 'dissolve', 'illumination_model')
    """
    Note: case-insensitive when parsing.

    Kd: diffuse color coefficient
    Ka: Ambient color coefficient
    Ks: Specular color coefficient
    d: Dissolve factor
    Ni: Refraction index
    Ns: Specular exponent, also known as glossiness, defines the focus of specular highlights in the material.
        Ns values normally range from 0 to 1000, with a high value resulting in a tight, concentrated highlight.
    illum:
        0 - disable lighting
        1 - ambient & diffuse (set specular color to black),
        2 - full lighting
    map_Kd: Diffuse color texture map
    map_Ks: Specular color texture map
    map_Ka: Ambient color texture map
    map_Ns: Shininess texture map
    map_Bump/map_bump/bump: Bump map
    map_Norm/map_norm/norm: Normal map
    map_Disp/map_disp/disp: Displacement map
    map_d: Opacity texture map
    refl: reflection map
    """

    def __init__(self, name,
                 diffuse_map: Texture = None, bump_map: Texture = None, normal_map: Texture = None,
                 ambient=(0.8, 0.8, 0.8), diffuse=(0.8, 0.8, 0.8), specular=(1.0, 1.0, 1.0),
                 shininess=1.0, ior=1.0, dissolve=1.0, illum=2):
        self.name = name  # material name
        self.ambient = np.asarray(ambient, dtype=np.float32)  # Ka
        self.diffuse = np.asarray(diffuse, dtype=np.float32)  # Kd
        self.specular = np.asarray(specular, dtype=np.float32)  # Ks

        self.diffuse_map = diffuse_map
        self.bump_map = bump_map
        self.normal_map = normal_map

        self.shininess = shininess  # Ns
        self.refractive_index = ior  # Ni
        self.dissolve = dissolve  # d
        self.illumination_model = illum  # illumination model

    def __repr__(self):
        return 'Material <{}>\n' \
               '  - diffuse_map: {}\n' \
               '  - bump_map: {}\n' \
               '  - normal_map: {}\n' \
               '  - ambient_color: {}\n' \
               '  - diffuse_color: {}\n' \
               '  - specular_color: {}\n' \
               '  - glossiness: {}\n' \
               '  - ior: {}\n' \
               '  - dissolve: {}\n'.format(self.name,
                                           self.diffuse_map.name,
                                           self.bump_map.name,
                                           self.normal_map.name,
                                           self.ambient,
                                           self.diffuse,
                                           self.specular,
                                           self.shininess,
                                           self.refractive_index,
                                           self.dissolve)


class Material:
    __slots__ = ('_data',)

    def __init__(self, name, ka=(0.6, 0.6, 0.6), kd=(0.8, 0.8, 0.8), ks=(0.8, 0.8, 0.8), shininess=1.0,
                 diffuse_map_path: str = None, normal_map_path: str = None, bump_map_path: str = None, **kwargs):
        from ..assets import default_asset_mgr as mgr
        self._data: MaterialData = mgr.get_or_create_material(name=name, ambient=ka, diffuse=kd, specular=ks,
                                                              shininess=shininess,
                                                              diffuse_map_path=diffuse_map_path,
                                                              normal_map_path=normal_map_path,
                                                              bump_map_path=bump_map_path, **kwargs)

    @property
    def inner(self):
        return self._data

    def __getattr__(self, item):
        return getattr(self._data, item)

    @property
    def name(self):
        return self._data.name



import os.path
import re

from ...assets import default_resolver

OBJ_MTL_TEXTURES = ['map_Kd', 'map_Ka', 'map_Ks', 'map_Ns', 'map_d', 'decal', 'refl',
                    'norm', 'map_norm', 'normal', 'map_normal', 'map_Norm', 'map_Normal', 'Norm', 'Normal',
                    'disp', 'map_disp', 'map_Disp', 'Disp',
                    'bump', 'map_bump', 'map_Bump', 'Bump']


def remove_comments(comment_symbol, text):
    return re.sub(f'{comment_symbol}.*', '', text).strip()


class PathResolver:
    def __init__(self, filepath):
        self._base = os.path.dirname(os.path.abspath(filepath))

    def resolve(self, filepath) -> str:
        final_path = filepath

        if not os.path.isabs(filepath):
            final_path = os.path.abspath(os.path.join(self._base, filepath))

        return final_path


class WavefrontReader:
    def __init__(self, filepath, resolver=default_resolver):
        self._filepath = resolver.resolve(filepath)

    def read(self):
        with open(self._filepath) as file:
            return self._load_obj(file, resolver=PathResolver(self._filepath))

    @staticmethod
    def _load_obj(file, resolver=None, split_object=False, group_material=True, skip_materials=False):
        """
        Load a Wavefront OBJ file (only contain triangles).

        Only accepts objects with V/VT/VN

        Args:
            file (file like object):
                Contains OBJ data.

            resolver (PathResolver):
                Resolve the path related while loading the OBJ file.

            split_object (bool):
                Split meshes at each `o` declared in file.

            group_material (bool):
                Group faces that share the same material into the same mesh.

            skip_materials (bool):
                Don't load any materials.

        Returns:
            dict with following format:
            {
                # data are stored following the order of objects.
                'vertices': [],
                'texcoords': [],
                'normals': [],
                'indices': [],  # vertex index to draw the object as whole
                'faces': [
                    [
                        (0, 1, 2),  # vertex index
                        (0, 1, 2),  # vertex tex coord idx
                        (0, 1, 2),  # vertex normal idx
                    ],
                    ...
                ],
                'materials': [],
                'objects': [
                    {
                        'name: 'object_000'
                        'vertex_range': (start, end),
                        'index_range': (start, end),
                        'texcoord_range': (start, end),
                        'normal_range': (start, end),
                        'face_range': (start, end)
                        'materials': [
                            { 'name': 'material_name_000', f_range=[face_idx_start, face_idx_end] },
                            { 'name': 'material_name_001', f_range=[face_idx_start, face_idx_end] }
                        ],
                        'vertex_format': 'V', 'V_T', 'V_N', 'V_T_N'
                    },
                    {
                        'name': 'object_001'
                        ...
                    },
                ]
            }
        """
        positions = []
        texcoords = []
        indices = []
        normals = []
        faces = []
        objects = []

        content = re.sub('#.*', '', file.read()).strip()  # remove the comments

        materials = {}

        for matched in re.finditer(r'mtllib\s(?P<mtl>.*)', content):
            if resolver:
                mtl_path = resolver.resolve(matched.group('mtl'))
            else:
                mtl_path = os.path.abspath(matched.group('mtl'))

            if os.path.exists(mtl_path):
                materials = WavefrontReader._parse_materials(mtl_path, resolver)
            else:
                print(f'Material file {mtl_path} not exists.')

        default_object = 'default_object_{}'

        c_obj_idx = -1
        c_mtl_idx = -1

        for line in content.splitlines():
            # split by whitespace
            line = line.split()
            # skip blank line
            if len(line) == 0:
                continue

            key, values_str = line[0], line[1:]

            diff = lambda arr: arr[1] - arr[0]

            if key in ('o', 'g'):
                c_obj_idx += 1
                # parse object name.
                obj_name = '_'.join(values_str)
                # in case of an empty name get from the default object name
                obj_name = obj_name if len(obj_name) > 0 else default_object.format(c_obj_idx)
                # initialise a new object
                objects.append({
                    'name': f'{key}_{obj_name}',
                    'vertex_range': [len(positions), -1],
                    'index_range': [len(indices) * 3, -1],
                    'texcoord_range': [len(texcoords), -1],
                    'normal_range': [len(normals), -1],
                    'face_range': [len(faces), -1],
                    'materials': [],
                    'vertex_format': ''
                })

                # save information for last parsed object before start reading next one
                if c_obj_idx >= 1:
                    objects[c_obj_idx - 1]['vertex_range'][1] = len(positions)
                    objects[c_obj_idx - 1]['index_range'][1] = len(indices) * 3
                    objects[c_obj_idx - 1]['texcoord_range'][1] = len(texcoords)
                    objects[c_obj_idx - 1]['normal_range'][1] = len(normals)
                    objects[c_obj_idx - 1]['face_range'][1] = len(faces)

                    # if the current object doesn't have a material
                    if len(objects[c_obj_idx - 1]['materials']) == 0:
                        objects[c_obj_idx - 1]['materials'].append({
                            'name': 'default_material',
                            'f_range': objects[c_obj_idx - 1]['face_range']
                        })

                    objects[c_obj_idx - 1]['materials'][c_mtl_idx]['f_range'][1] = len(faces)

                    vertex_format = 'V'

                    if diff(objects[c_obj_idx - 1]['texcoord_range']) > 0:
                        vertex_format += '_T'

                    if diff(objects[c_obj_idx - 1]['normal_range']) > 0:
                        vertex_format += '_N'

                    objects[c_obj_idx - 1]['vertex_format'] = vertex_format

                # reset material index
                c_mtl_idx = -1

            elif key == 'v':
                positions.append([float(x) for x in values_str])

            elif key == 'vt':
                texcoords.append([float(x) for x in values_str])

            elif key == 'vn':
                normals.append([float(x) for x in values_str])

            elif key == 'f':
                # parse line into a list
                parsed = [tuple(int(e) - 1 if e != '' else -1 for e in c.split('/')) for c in values_str]
                vertex_count = len(parsed)
                if vertex_count < 3:
                    # skip the face with less 3 vertices
                    continue
                else:
                    # in case of a face with more than 3 points, apply the simplest triangulation
                    for i in range(0, len(parsed) - 2):
                        faces.append(list(zip(*[parsed[0], *parsed[i + 1: i + 3]])))

            elif key == 'usemtl':
                c_mtl_idx += 1
                mtl_name = '_'.join(values_str)
                mtl_name = mtl_name if len(mtl_name) > 0 else default_object.format(c_obj_idx)
                objects[c_obj_idx]['materials'].append({
                    'name': mtl_name,
                    'f_range': [len(faces), -1]
                })

                # save information for last mtl before start reading next one
                if c_mtl_idx >= 1:
                    objects[c_obj_idx]['materials'][c_mtl_idx - 1]['f_range'][1] = len(faces)

            else:
                # ignoring s
                pass

        objects[c_obj_idx]['vertex_range'][1] = len(positions)
        objects[c_obj_idx]['index_range'][1] = len(indices)
        objects[c_obj_idx]['texcoord_range'][1] = len(texcoords)
        objects[c_obj_idx]['normal_range'][1] = len(normals)
        objects[c_obj_idx]['face_range'][1] = len(faces)
        objects[c_obj_idx]['materials'][c_mtl_idx]['f_range'][1] = len(faces)

        vertex_format = 'V'

        if diff(objects[c_obj_idx - 1]['texcoord_range']) > 0:
            vertex_format += '_T'

        if diff(objects[c_obj_idx - 1]['normal_range']) > 0:
            vertex_format += '_N'

        objects[c_obj_idx]['vertex_format'] = vertex_format

        # Deal with the case `group` is used after `object` right before the usemtl statement
        # instead of using the usemtl statement directly after the object statement.
        # o object_name
        # v ...
        # vt ...
        # vn ...
        # g group_name
        # usemtl material_name
        # f ...
        tobe_removed = []
        if len(objects) > 1:
            for i in range(0, len(objects), 2):
                first = objects[i]
                if i + 1 < len(objects):
                    second = objects[i + 1]
                    if second['vertex_range'][0] == second['vertex_range'][1] and \
                            second['index_range'][0] == second['index_range'][1] and \
                            second['texcoord_range'][0] == second['texcoord_range'][1] and \
                            second['normal_range'][0] == second['normal_range'][1] and \
                            first['vertex_range'][1] == second['vertex_range'][0] and \
                            first['index_range'][1] == second['index_range'][0] and \
                            first['texcoord_range'][1] == second['texcoord_range'][0] and \
                            first['normal_range'][1] == second['normal_range'][0]:
                        second['vertex_range'][0] = first['vertex_range'][0]
                        second['index_range'][0] = first['index_range'][0]
                        second['texcoord_range'][0] = first['texcoord_range'][0]
                        second['normal_range'][0] = first['normal_range'][0]
                        tobe_removed.append(i)
                    for j in tobe_removed:
                        objects.pop(j)

        return {
            'positions': positions,
            'texcoords': texcoords,
            'normals': normals,
            'indices': indices,
            'faces': faces,
            'materials': materials,
            'objects': objects,
        }

    @staticmethod
    def _parse_materials(path, resolver: PathResolver = None) -> dict:
        """
        Parse a MTL file.

        Args:
            resolver (PathResolver):
                Help finding the correct file path to textured file (in needed).

            path (str):
                File path to mtl file.

        Returns:
            dict of materials indexed by its name:
            {
                'material_name_000': {
                    'Ns': ...,
                    'Ka': ...,
                    ...
                    'map_Kd': {
                        'params': {
                            ... (optional)  # non default parameters
                        },
                        'path': 'path/to/file'
                    }
                }
            }
        """
        materials = {}

        with open(path) as file:
            lines = remove_comments('#', file.read()).strip().splitlines()

            current_material_name = None
            # parse mtl file line by line
            for line in lines:
                components = line.strip().split()
                if len(components) <= 1:
                    continue
                key = components[0]
                if key == 'newmtl':
                    # new material
                    current_material_name = '_'.join(components[1:])
                    materials[current_material_name] = {}
                elif key in OBJ_MTL_TEXTURES:
                    # textures
                    index = OBJ_MTL_TEXTURES.index(key)
                    key_name = key if index < 7 else (
                        'map_norm' if index <= 14 else ('map_disp' if index <= 18 else 'map_bump'))
                    materials[current_material_name][key_name] = WavefrontReader._parse_texture(components[1:],
                                                                                                resolver)
                else:
                    # other properties: Kd, Ka, Ks, d, illum, Ns, Ni
                    value = [float(x) for x in components[1:]]
                    if len(value) == 1:
                        value = value[0]
                    materials[current_material_name][key] = value

            return materials

    @staticmethod
    def _parse_texture(line, resolver: PathResolver = None) -> dict:
        """
        Parse texture along with its parameters.

        Args:
            line: Line of texture parameters and path (already split).

        Returns:
            dict of texture parameters indexed by its name:
            {
                ... (optional)  # non default parameters
            },
            See `ObjTextureParams` for more details.
        """
        # parse texture path
        params_str, path_str = line[:-1], line[-1]
        texture_path = path_str if resolver is None else resolver.resolve(path_str)
        # parse texture parameters
        params_pos = [i for i, x in enumerate(params_str) if x.startswith('-')]
        params = {}
        for i, pos in enumerate(params_pos):
            param_name = params_str[pos][1:]
            # last parameter
            if i == len(params_pos) - 1:
                param_value = params_str[pos + 1:]
            else:
                param_value = params_str[pos + 1: params_pos[i + 1]]
            params[param_name] = param_value

        return {
            'path': texture_path,
            'params': params,
        }



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='bk7084',
    version='0.2.6',
    author='Yang Chen, Ruben Wiersma, Ricardo Marroquim',
    author_email="matthiasychen@gmail.com, rubenwiersma@gmail.com, R.Marroquim@tudelft.nl",
    description='Python framework for BK7084 Computational Simulations',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/bk7084/framework",
    project_urls={
        "Bug Tracker": "https://github.com/bk7084/framework/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(where=".", include=["bk7084", "bk7084.*"]),
    package_data={
        'bk7084': ['assets/models/*', 'assets/shaders/*', 'assets/textures/*']
    },
    python_requires='>=3.9,<3.11',
    install_requires=[
        'numpy',
        'pyopengl',
        'glfw',
        'pypng',
        'imgui[glfw]',
        'Pillow',
        'numba'
    ],
)



import os
import __main__


class PathResolver:
    def __init__(self, search_paths=None):
        _search_paths = [
            os.path.abspath(os.path.dirname(__file__)),
            os.path.abspath(os.path.curdir),
        ]
        if hasattr(__main__, '__file__'):
            _search_paths.insert(0, os.path.abspath(os.path.dirname(__main__.__file__)))

        if search_paths is not None:
            for path in search_paths:
                abs_path = os.path.abspath(path)
                if os.path.exists(abs_path):
                    _search_paths.append(abs_path)

        self._search_paths = list(set(_search_paths))

    def resolve(self, filepath: str):
        valid_path = None
        if os.path.isabs(filepath) and os.path.exists(filepath):
            valid_path = filepath
        else:
            for base_path in self._search_paths:
                path = os.path.join(base_path, filepath)
                if os.path.exists(path):
                    valid_path = path
                    break
        if valid_path is None:
            raise ValueError(f'Cannot resolve path: {filepath}')
        return valid_path


default_resolver = PathResolver()



from .. import gl

import enum
from collections import OrderedDict
from dataclasses import dataclass


@enum.unique
class VertexAttrib(enum.Enum):
    Position = 0,
    Normal = 1,
    Tangent = 2,
    BiTangent = 3,
    Color0 = 4,
    Color1 = 5,
    Color2 = 6,
    Color3 = 7,
    BlendIndices = 8,
    BlendWeight = 9,
    TexCoord0 = 10,
    TexCoord1 = 11,
    TexCoord2 = 12,
    TexCoord3 = 13,
    TexCoord4 = 14,
    TexCoord5 = 15,
    TexCoord6 = 16,
    TexCoord7 = 17,

    @property
    def default_dimension(self):
        if self is VertexAttrib.Position or self is VertexAttrib.Normal:
            return 3
        elif VertexAttrib.TexCoord0 <= self.value <= VertexAttrib.TexCoord7:
            return 2
        elif self is VertexAttrib.Tangent or VertexAttrib.Color0 <= self.value <= VertexAttrib.Color3:
            return 4
        else:
            raise Exception(f"{self.name} doesn't have default dimension.")


@enum.unique
class VertexAttribFormat(enum.Enum):
    Float64 = gl.GL_DOUBLE
    Float32 = gl.GL_FLOAT
    UInt32 = gl.GL_UNSIGNED_INT,
    SInt32 = gl.GL_INT,
    Float16 = gl.GL_HALF_FLOAT
    UInt16 = gl.GL_UNSIGNED_SHORT,
    SInt16 = gl.GL_SHORT,
    UInt8 = gl.GL_UNSIGNED_BYTE
    SInt8 = gl.GL_BYTE

    def size_in_bytes(self):
        if self is VertexAttribFormat.Float64:
            return 8
        elif self is VertexAttribFormat.Float32 or self is VertexAttribFormat.UInt32 or self is VertexAttribFormat.SInt32:
            return 4
        elif self is VertexAttribFormat.Float16 or self is VertexAttribFormat.UInt16 or \
                self is VertexAttribFormat.SInt16 or self is VertexAttribFormat.SNorm16:
            return 2
        else:
            return 1

    @property
    def gl_type(self):
        return self.value


@dataclass
class VertexAttribDescriptor:
    attrib: VertexAttrib
    format: VertexAttribFormat
    dimension: int

    def __init__(self, attrib: VertexAttrib, fmt: VertexAttribFormat, dim=1):
        self.attrib = attrib
        self.format = fmt
        self.dimension = dim


class VertexLayout:
    def __init__(self, *descriptors):
        """Construct the vertex layout from vertex attribute descriptors.

        Args:
            *descriptors:
                Vertex attribute descriptors specify the vertex attribute and its associated data type
                (format) and dimension. It could be either `VertexAttribDescriptor` objects or tuples
                of form:

                (attrib, format, n),

                `attrib` is the `VertexAttrib`, `format` is the `VertexAttribFormat`. `n` is the dimension of this attribute.
        """
        self._description = OrderedDict()
        self._stride = 0
        if descriptors:
            for desc in descriptors:
                if isinstance(desc, VertexAttribDescriptor):
                    self._description[desc.attrib] = (desc.format, desc.dimension)
                    self._stride += desc.format.size_in_bytes() * desc.dimension
                elif isinstance(desc, tuple):
                    self._description[desc[0]] = (desc[1], desc[2])
                    self._stride += desc[1].size_in_bytes() * desc[2]

    @classmethod
    def new(cls):
        return cls()

    @property
    def description(self):
        return self._description

    @classmethod
    def default(cls):
        return cls(VertexAttribDescriptor(VertexAttrib.Position, VertexAttribFormat.Float32, 3),
                   VertexAttribDescriptor(VertexAttrib.Color0, VertexAttribFormat.Float32, 4),
                   VertexAttribDescriptor(VertexAttrib.TexCoord0, VertexAttribFormat.Float32, 2),
                   VertexAttribDescriptor(VertexAttrib.Normal, VertexAttribFormat.Float32, 3),
                   VertexAttribDescriptor(VertexAttrib.Tangent, VertexAttribFormat.Float32, 3))

    def has(self, attrib: VertexAttrib):
        return attrib in self._description

    def offset_of(self, attrib: VertexAttrib):
        if attrib not in self._description:
            raise ValueError(f'{attrib} not exists in current vertex layout.')

        offset = 0
        for attr, (fmt, dim) in self._description.items():
            if attr is attrib:
                return offset
            else:
                offset += fmt.size_in_bytes() * dim

        return offset

    @property
    def stride(self):
        """Returns vertex stride (size of a single vertex in bytes)."""
        return self._stride

    @property
    def attrib_count(self):
        return len(self._description)

    def compute_buffer_size(self, count):
        """Returns the size of vertex buffer in bytes for a certain count of vertices.

        Args:
            count (int): Numbers of vertices.

        Returns:
            int. Size of vertex buffer for a specific number of vertices.
        """
        return count * self._stride



import ctypes
import logging
from typing import Union

import numpy as np

from .array import VertexArrayObject
from .buffer import VertexBuffer, IndexBuffer
from .vertex_layout import VertexLayout, VertexAttrib, VertexAttribFormat
from .. import gl
from .. import app
from ..math import Mat4, Vec3
from ..scene import Mesh


def draw(*objs, **kwargs):
    """Draws a shape object."""
    # record shape and its associated vbo. avoid to create multiple vertex buffer object
    # for the same object each time the function is called.
    shader = kwargs.get('shader', None)
    shader_params = kwargs.get('shader_params', None)
    transform = kwargs.get('transform', Mat4.identity())

    if not hasattr(draw, 'shapes_created_gpu_objects'):
        draw.shapes_created_gpu_objects = {}

    from ..geometry.shape import Shape
    from ..scene.mesh import Mesh

    for obj in objs:
        if isinstance(obj, Shape):
            if obj not in draw.shapes_created_gpu_objects:
                vbo = VertexBuffer(obj.vertex_count,
                                   VertexLayout((VertexAttrib.Position, VertexAttribFormat.Float32, 3),
                                                (VertexAttrib.Color0, VertexAttribFormat.Float32, 4)))
                vbo.set_data(obj.interleaved_vertices)

                ibo = IndexBuffer(obj.index_count)
                ibo.set_data(obj.indices.astype(np.uint32))

                vao = VertexArrayObject()

                vao.bind_vertex_buffer(vbo, [0, 1])

                draw.shapes_created_gpu_objects[obj] = (vao, vbo, ibo)

            vao, vbo, ibo = draw.shapes_created_gpu_objects[obj]

            if obj.is_dirty:
                vbo.set_data(obj.interleaved_vertices)
                ibo.set_data(obj.indices.astype(np.uint32))
                obj.is_dirty = False

            shader = app.current_window().default_shader if shader is None else shader
            with shader:
                shader.model_mat = transform
                shader.shading_enabled = False
                shader['mtl.enabled'] = False
                with vao:
                    with ibo:
                        gl.glDrawElements(obj.drawing_mode.value, ibo.index_count, gl.GL_UNSIGNED_INT,
                                          ctypes.c_void_p(0))

        elif isinstance(obj, Mesh):
            obj.draw(shader=shader, shader_params=shader_params)

        else:
            logging.info('Nothing to draw.')




"""Framework written in Python for BK7084 Computational Simulations"""

from . import app
from . import camera
from . import geometry
from . import graphics
from .app.window import *
from .app import ui
from .app.input import *
from . import math

__version__ = (0, 2, 6)



import abc
import math

from ...math import Vec3
from ...misc import Color


class Light(metaclass=abc.ABCMeta):
    @classmethod
    def __subclasshook__(cls, subclass):
        return hasattr(subclass, 'matrix')

    def __init__(self, position: Vec3, color: Color, **kwargs):
        """

        Args:
            direction:
            color:
            **kwargs:
                sm_width (int): shadow map width
                sm_height (int): shadow map height
                sm_near (float): shadow map near plane
                sm_far (float): shadow map far plane
                sm_fov (float): shadow map vertical field of view (in degrees or radians).
        """
        self._position = position
        self._color = color
        self._is_dirty = True

        # Shadow map related variables
        self._sm_width = kwargs.get('sm_width', 1024)
        self._sm_height = kwargs.get('sm_width', 1024)
        self._sm_near = kwargs.get('sm_near', 0.1)
        self._sm_far = kwargs.get('sm_far', 100.0)
        fov = kwargs.get('sm_fov', 60.0)
        is_degrees = kwargs.get('degrees', True)
        self._sm_fov = math.radians(fov) if is_degrees else fov
        self._sm_is_perspective = kwargs.get('is_perspective', False)

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, value: Vec3):
        self._position = value
        self._is_dirty = True

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, value: Color):
        self._color = value

    @property
    @abc.abstractmethod
    def matrix(self):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def view_matrix(self):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def is_perspective(self):
        raise NotImplementedError

    @property
    def is_directional(self):
        return NotImplementedError



__all__ = [
    'AreaLight',
    'DirectionalLight',
    'PointLight',
    'SpotLight'
]

from .area_light import AreaLight
from .directional_light import DirectionalLight
from .point_light import PointLight
from .spot_light import SpotLight



import itertools

# from bk7084 import app
from .. import app
from ..math import Vec4
from numba import njit, prange

from .building import Component, Building
from .mesh import Mesh
from .entity import Entity
from .. import gl
from ..app.input import KeyCode, KeyModifier
from ..app import ui
from ..assets import default_asset_mgr
from ..camera import Camera
from ..graphics.lights.light import Light
from ..graphics.framebuffer import Framebuffer
from ..graphics.lights import PointLight, DirectionalLight
from ..misc import PaletteDefault
from ..math import Mat4, Vec3


class MeshEntity(Entity):
    def __init__(self, mesh: Mesh, cast_shadow=True):
        super().__init__()
        self.mesh = mesh

    def draw(self, shader=None, **kwargs):
        """
        Args:
            shader (ShaderProgram): Shader to use. (Override the shader of mesh)
            **kwargs: Extra uniforms that are going to be passed to shader
        """
        self.mesh.draw(shader=shader, **kwargs)

    @property
    def meshes(self):
        return [(self.mesh, Mat4.identity())]


class Scene:
    """
    Scene manages the rendering of meshes, lights and cameras.
    """
    def __init__(self, window, entities=(), camera=None, num_lights=12, draw_light=False,
                 bg_color=PaletteDefault.Background.as_color(), **kwargs):
        """
        Initialisation of a Scene object.

        Args:
            window:
                Specifies the window in which the scene will be presented

            entities (list of Meshes or instances of subclasses of Entity):
                All entity instances that you want to have in the Scene

            camera (Camera):
                Main camera of the scene.

            light (Light):
                Light inside

            bg_color (Color): screen clear color
            **kwargs:
        """
        self._window = window
        self._entities = []
        if len(entities) > 0:
            for entity in entities:
                if isinstance(entity, Mesh):
                    self._entities.append(MeshEntity(entity))
                elif isinstance(entity, Entity):
                    self._entities.append(entity)
                elif isinstance(entity, Component):
                    self._entities.append(entity)

        self._lights = []
        delta = 180.0 / num_lights
        for i in range(num_lights):
            position = Mat4.from_rotation_z(delta * i, True) * Vec4(15.0, 1.0, 15.0, 1.0)
            self._lights.append(PointLight(Vec3(position)))
        self._current_light = 0

        self._draw_light = draw_light
        self._light_sphere = Mesh('light_sphere', 'models/uv_sphere.obj', texture_enabled=False,
                                  init_transform=Mat4.from_scale(Vec3(0.1, 0.1, 0.1)))
        self._cameras = []
        self._main_camera = -1
        if camera is not None:
            self._cameras.append(camera)
            self._main_camera = 0
            self._window.attach_listeners(camera)

        self._window.attach_listeners(self)
        self._values = 1.0, 2.0, 3.0
        self._depth_map_framebuffer = Framebuffer(self._window.width, self._window.height, depth_shader_accessible=True)
        self._depth_map_pipeline = default_asset_mgr.get_or_create_pipeline('depth_map',
                                                                            vertex_shader='shaders/depth_map.vert',
                                                                            pixel_shader='shaders/depth_map.frag')
        self._energy_framebuffer = Framebuffer(self._window.width, self._window.height)
        self._energy_pipeline = default_asset_mgr.get_or_create_pipeline('energy_pipeline',
                                                                         vertex_shader='shaders/energy_map.vert',
                                                                         pixel_shader='shaders/energy_map.frag')
        self._is_wire_frame = False
        self._rendering_info = self.extract_rendering_info()

    def set_main_camera(self, index):
        """
        Specifies the camera going to be used as the main camera.

        Args:
            index (int): index of cameras in the scene.

        Returns:
            None
        """
        if index != self._main_camera:
            self._switch_to_camera(index)

    @property
    def lights(self):
        return self._lights

    @property
    def current_light(self):
        return self._lights[self._current_light]

    @property
    def main_camera(self):
        return self._cameras[self._main_camera]

    def create_camera(self, pos, look_at, up, fov_v=45.0, near=0.1, far=1000., degrees=True, zoom_enabled=False, safe_rotations=True) -> int:
        """
        Creates a new camera and return its index inside the scene.

        Args:
            pos (array like):
                Specifies the position of the camera

            look_at (array like):
                Specifies look at of the camera

            up (array like):
                Specifies up-vector of the camera

            fov_v (float):
                Specifies vertical field of view of the camera

            near (float):
                Specifies near clip plane

            far (float):
                Specifies far clip plane

            degrees (bool):
                Specifies how to treat input angles

            zoom_enabled (bool):
                Specifies whether to enable camera zoom or not

            safe_rotations:
                Specifies whether to enable safe rotations by disable certain cases

        Returns:
            int, index of the newly created camera inside Scene
        """
        camera = Camera(pos, look_at, up, self._window.width / self._window.height, fov_v, near, far, degrees, zoom_enabled, safe_rotations)
        camera.on_resize(self._window.width, self._window.height)
        self._cameras.append(camera)
        if len(self._cameras) == 1:
            self._main_camera = 0
            self._window.attach_listeners(self._cameras[0])

        return len(self._cameras) - 1

    def _switch_to_camera(self, to_idx):
        idx = self._main_camera
        self._window.detach_listeners(self._cameras[idx])
        self._main_camera = to_idx
        self._window.attach_listeners(self._cameras[to_idx])
        self._cameras[to_idx].on_resize(self._window.width, self._window.height)
        
    @property
    def depth_map_framebuffer(self):
        return self._depth_map_framebuffer

    def extract_rendering_info(self):
        # True --> cast shadow
        # False --> do not cast shadow
        info = {
            True: {},
            False: {},
        }
        for entity in self._entities:
            cast_shadow = entity.cast_shadow
            if entity.drawable:
                for mesh, transform in entity.meshes:
                    for pipeline_uuid, records in mesh.rendering_info.items():
                        if pipeline_uuid not in info[cast_shadow]:
                            info[cast_shadow][pipeline_uuid] = []
                        for record in records:
                            info[cast_shadow][pipeline_uuid].append((*record, transform))
        return info

    def load_mesh_entity(self, name, filename):
        assert filename is not None and name is not None
        entity = MeshEntity(Mesh(name, filename))
        self._entities.append(entity)
        return entity

    def on_gui(self):
        if ui.tree_node('Light'):
            if self._lights[self._current_light].is_directional:
                ui.drag_float3('Position', *self._lights[0].position)
            else:
                _, self._lights[0].position = ui.drag_float3('Position', *self._lights[0].position)
            _, self._lights[0].color.rgb = ui.color_edit3('Color', *self._lights[0].color.rgb)
            if self._lights[0].is_directional:
                _, self._lights[0].direction = ui.drag_float3('Direction', *self._lights[0].direction)
            if ui.button('Next light'):
                self._current_light += 1
                self._current_light %= len(self._lights)
            ui.tree_pop()

        if ui.tree_node('Camera'):
            if ui.button("Next Camera"):
                if len(self._cameras) > 0:
                    self._switch_to_camera((self._main_camera + 1) % len(self._cameras))
                    print(self._main_camera)
            ui.tree_pop()

        if ui.tree_node('Rendering'):
            _, self._is_wire_frame = ui.checkbox('Wireframe', self._is_wire_frame)
            ui.tree_pop()

    def on_update(self, dt):
        self._rendering_info = self.extract_rendering_info()

    def on_key_press(self, key, mods):
        if key == KeyCode.D and mods == KeyModifier.Shift:
            main_cam = self._cameras[self._main_camera]
            self._depth_map_framebuffer.save_depth_attachment(main_cam.near, main_cam.far, self._lights[0].is_perspective)

        if key == KeyCode.C and mods == KeyModifier.Shift:
            self._depth_map_framebuffer.save_color_attachment()

    @staticmethod
    def _compute_energy(energy_map):
        energy_map = energy_map.reshape((-1, 4))
        visible_count, occluded_count, received_energy = _calc_energy(energy_map)
        visibility_ratio = visible_count / (occluded_count + visible_count)
        energy_ratio = received_energy / (occluded_count + visible_count)
        # print('occluded pixel count: ', occluded_count)
        # print('visible pixel count: ', visible_count)
        # print('visibility ratio: ', visibility_ratio)
        # print('energy ratio: ', energy_ratio)
        # print('energy: ', received_energy)
        return energy_ratio

    def energy_of_building_component(self, building: Building, comp: Component, light: Light = None, save_energy_map=False):
        light = self._lights[self._current_light] if light is None else light
        with self._energy_framebuffer:
            self._energy_framebuffer.enable_depth_test()
            self._energy_framebuffer.clear((1.0, 1.0, 1.0, 1.0))
            comp.compute_energy(self._energy_pipeline, light, self._window.size,
                                self._depth_map_framebuffer.depth_attachments[0],
                                [building.transform_of(comp)])

        energy_map = self._energy_framebuffer.save_color_attachment(save_as_image=save_energy_map)
        return self._compute_energy(energy_map)

    def energy_of_building(self, building: Building, light: Light = None, save_energy_map=False, mesh_transform=Mat4.identity()):
        light = self._lights[self._current_light] if light is None else light

        with self._energy_framebuffer:
            self._energy_framebuffer.enable_depth_test()
            self._energy_framebuffer.clear((1.0, 1.0, 1.0, 1.0))
            if isinstance(building, Building):
                building.compute_energy(self._energy_pipeline, light,
                                        self._window.size,
                                        self._depth_map_framebuffer.depth_attachments[0])
            elif isinstance(building, Mesh):
                building.compute_energy(self._energy_pipeline, mesh_transform, light, self._window.size,
                                        self._depth_map_framebuffer.depth_attachments[0])
        energy_map = self._energy_framebuffer.save_color_attachment(save_as_image=save_energy_map)
        return self._compute_energy(energy_map)

    def draw_v2(self, shader=None, auto_shadow=False, **kwargs):
        light = self._lights[self._current_light]
        camera = self._cameras[self._main_camera]
        old_polygon_mode = gl.glGetIntegerv(gl.GL_POLYGON_MODE)[0]

        if auto_shadow:
            # two passes
            # 1st pass
            if old_polygon_mode != gl.GL_LINE:
                with self._depth_map_framebuffer:
                    gl.glEnable(gl.GL_DEPTH_TEST)
                    self._depth_map_framebuffer.clear(PaletteDefault.Background.as_color().rgba)
                    with self._depth_map_pipeline:
                        for _, records in self._rendering_info[True].items():
                            self._bind_light(self._depth_map_pipeline, light)
                            for record in records:
                                vao, topology, count, _, _, _, _, _, _, _, _, mesh_transform, transform = record
                                self._depth_map_pipeline['model_mat'] = transform * mesh_transform
                                with vao:
                                    gl.glDrawArrays(topology, 0, count)
            # second pass
            gl.glClearColor(*PaletteDefault.BlueA.as_color().rgba)
            gl.glEnable(gl.GL_DEPTH_TEST)
            gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
            if self._is_wire_frame:
                gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_LINE)
            else:
                gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_FILL)

            for cast_shadow, info in self._rendering_info.items():
                for pipeline_uuid, records in info.items():
                    pipeline = default_asset_mgr.get_pipeline(pipeline_uuid)
                    with pipeline:
                        pipeline['time'] = app.current_window().elapsed_time
                        self._bind_light(pipeline, light)
                        self._bind_camera(pipeline, camera)
                        for record in records:
                            vao, topology, count, mtl, diffuse_map, shading_enabled, \
                            mtl_enabled, tex_enabled, normal_map_enabled, bump_map_enabled, \
                            parallax_map_enabled, mesh_transform, transform = record
                            pipeline['shading_enabled'] = shading_enabled
                            pipeline['model_mat'] = transform * mesh_transform
                            if cast_shadow:
                                pipeline['shadow_map_enabled'] = True
                            else:
                                pipeline['shadow_map_enabled'] = True
                            self._bind_material(pipeline, mtl, mtl_enabled, tex_enabled, normal_map_enabled,
                                                bump_map_enabled, parallax_map_enabled)
                            with vao:
                                pipeline.active_texture_unit(0)
                                with diffuse_map:
                                    pipeline.active_texture_unit(1)
                                    with mtl.bump_map:
                                        pipeline.active_texture_unit(2)
                                        with mtl.normal_map:
                                            pipeline.active_texture_unit(3)
                                            with self._depth_map_framebuffer.depth_attachments[0]:
                                                gl.glDrawArrays(topology, 0, count)

            gl.glPolygonMode(gl.GL_FRONT_AND_BACK, old_polygon_mode)
        else:
            # single pass
            if shader is not None:
                with shader:
                    self._bind_light(shader, light)
                    self._bind_camera(shader, camera)

            if self._is_wire_frame:
                gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_LINE)
            else:
                gl.glPolygonMode(gl.GL_FRONT_AND_BACK, gl.GL_FILL)

            for pipeline_uuid, records in itertools.chain(self._rendering_info[True].items(), self._rendering_info[False].items()):
                pipeline = default_asset_mgr.get_pipeline(pipeline_uuid)
                with pipeline:
                    pipeline['time'] = app.current_window().elapsed_time
                    self._bind_light(pipeline, light)
                    self._bind_camera(pipeline, camera)
                    for record in records:
                        vao, topology, count, mtl, diffuse_map, shading_enabled, \
                        mtl_enabled, tex_enabled, normal_map_enabled, bump_map_enabled,\
                        parallax_map_enabled, mesh_transform, transform = record
                        pipeline['shading_enabled'] = shading_enabled
                        pipeline['model_mat'] = transform * mesh_transform
                        self._bind_material(pipeline, mtl, mtl_enabled, tex_enabled, normal_map_enabled, bump_map_enabled, parallax_map_enabled)
                        with vao:
                            pipeline.active_texture_unit(0)
                            with diffuse_map:
                                pipeline.active_texture_unit(1)
                                with mtl.bump_map:
                                    pipeline.active_texture_unit(2)
                                    with mtl.normal_map:
                                        gl.glDrawArrays(topology, 0, count)
            gl.glPolygonMode(gl.GL_FRONT_AND_BACK, old_polygon_mode)

        if self._draw_light:
            self._light_sphere.transform = Mat4.from_translation(self._lights[self._current_light].position)
            self._light_sphere.draw(camera=self._cameras[self._main_camera])

    def _bind_light(self, pipeline, light):
        pipeline['in_light_pos'] = light.position
        pipeline['in_light_dir'] = light.direction if light.is_directional else Vec3(0.0)
        pipeline['is_directional'] = light.is_directional
        pipeline['light_color'] = light.color.rgb
        pipeline['light_mat'] = light.matrix
        pipeline['near'] = light._sm_near
        pipeline['far'] = light._sm_far
        pipeline['is_persepective'] = light._sm_is_perspective

    def _bind_material(self, pipeline, mtl, mtl_enabled, tex_enabled, normal_map_enabled, bump_map_enabled, parallax_map_enabled):
        pipeline['mtl.diffuse'] = mtl.diffuse
        pipeline['mtl.ambient'] = mtl.ambient
        pipeline['mtl.specular'] = mtl.specular
        pipeline['mtl.shininess'] = mtl.shininess
        pipeline['mtl.enabled'] = mtl_enabled
        pipeline['mtl.use_diffuse_map'] = tex_enabled
        pipeline['mtl.use_normal_map'] = normal_map_enabled
        pipeline['mtl.use_bump_map'] = bump_map_enabled
        pipeline['mtl.use_parallax_map'] = parallax_map_enabled

        pipeline['mtl.diffuse_map'] = 0
        pipeline['mtl.bump_map'] = 1
        pipeline['mtl.normal_map'] = 2
        pipeline['shadow_map'] = 3

    def _bind_camera(self, pipeline, camera):
        pipeline['view_mat'] = camera.view_matrix
        pipeline['proj_mat'] = camera.projection_matrix


@njit(parallel=True, fastmath=True)
def _calc_energy(data):
    occluded_count = 0
    visible_count = 0
    received_energy = 0.0
    for i in prange(data.shape[0]):  # iterate over rows
        if data[i][0] == 255 and data[i][1] == 0 and data[i][2] == 0:
            occluded_count += 1
        if data[i][0] == 0 and data[i][2] == 0:
            visible_count += 1
            received_energy += data[i][1] / 255.0
    return visible_count, occluded_count, received_energy



import abc
from re import S
import uuid
from collections import namedtuple

from .mesh import Mesh, SubMesh
from .entity import Entity
from ..math import Mat4
from ..misc import PaletteDefault as Palette

import numpy as np


class Component(Entity):
    def __init__(self, name=None, cast_shadow=True):
        super(Component, self).__init__(name, cast_shadow)
        self._transform = Mat4.identity()

    @property
    @abc.abstractmethod
    def mesh(self) -> Mesh:
        raise NotImplementedError

    @property
    def transform(self):
        return self._transform

    @transform.setter
    def transform(self, value: Mat4):
        self._transform = value

    @property
    def cast_shadow(self):
        return self._cast_shadow

    @cast_shadow.setter
    def cast_shadow(self, value):
        self._cast_shadow = value

    @property
    def drawable(self):
        return self._is_drawable

    @drawable.setter
    def drawable(self, value):
        self._is_drawable = value

    @property
    def meshes(self):
        return [(self.mesh, self._transform)]

    def draw(self, matrices=None, **kwargs):
        if self._is_drawable:
            matrix = Mat4.identity()
            if matrices is not None:
                for m in matrices:
                    matrix = m * matrix
            self.mesh.draw(matrix, **kwargs)

    def compute_energy(self, shader, light, viewport_size, depth_map, matrices=None):
        if self._is_drawable:
            transform = Mat4.identity()
            if matrices is not None:
                for m in matrices:
                    transform = m * transform
            self.mesh.compute_energy(shader, transform, light, viewport_size, depth_map)


class Building(Entity):
    def __init__(self, name=None):
        super().__init__(name)
        self._components = []  # stores the Component objects
        self._root_components = []  # store the Component objects index
        self._hierarchy = {}  # store the parent index of components: { comp_index: parent_index }
        self._transform = Mat4.identity()

    def append(self, comp: Component, parent: Component = None):
        exists = False
        for c in self._components:
            if c.uuid == comp.uuid:
                exists = True
                break

        if exists:
            raise ValueError('Component already exists.')

        self._components.append(comp)
        index = len(self._components) - 1

        parent_index = -1

        if parent is not None:
            try:
                parent_index = self._components.index(parent)
            except ValueError:
                raise ValueError('Parent does not exist.')

        self._hierarchy[index] = parent_index

        if parent_index == -1:
            self._root_components.append(index)

    @property
    def components(self):
        return self._components

    @property
    def transform(self):
        return self._transform

    @transform.setter
    def transform(self, value: Mat4):
        self._transform = value

    def _parent_list(self, comp, parent_list):
        if comp in self._root_components:
            return parent_list
        else:
            if comp in self._hierarchy:
                parent_list.append(self._hierarchy[comp])
                return self._parent_list(self._hierarchy[comp], parent_list)
            else:
                return []

    def transform_of(self, comp):
        matrix = Mat4.identity()
        if comp in self._components:
            idx = self._components.index(comp)
            parents = self._parent_list(idx, [idx])
            matrices = [self._components[p].transform for p in parents] + [self.transform]
            for m in matrices:
                matrix = m * matrix
        return matrix

    def draw(self, shader=None, **kwargs):
        for idx, comp in enumerate(self._components):
            if comp.drawable:
                parents = self._parent_list(idx, [idx])
                matrices = [self._components[p].transform for p in parents] + [self.transform]
                comp.draw(matrices, shader=shader, **kwargs)

    def compute_energy(self, shader, light, viewport_size, depth_map):
        for idx, comp in enumerate(self._components):
            if comp.drawable:
                parents = self._parent_list(idx, [idx])
                matrices = [self._components[p].transform for p in parents] + [self.transform]
                comp.compute_energy(shader, light, viewport_size, depth_map, matrices)

    def convert_to_mesh(self):
        """ Converts a building to a single mesh that can be rendered more quickly.
        """
        print(f"Converting building to mesh...{self.name}")
        if len(self._components) == 0:
            return None
            
        vertices_dict, vertex_idx = {}, 0
        uvs_dict, uvs_idx = {}, 0
        normals_dict, normal_idx = {}, 0
        triangles, tri_offset = [], 0
        sub_mesh_dict = {}
        for comp in self._components:
            if comp.drawable:
                transform = np.array(self.transform_of(comp))
                mesh = comp.mesh
                
                # Transform vertices with model matrix
                comp_vertices = (transform @ np.concatenate((mesh.vertices, np.ones_like(mesh.vertices[:, 0:1])), axis=1).T).T
                comp_vertices = comp_vertices[:, :3] / comp_vertices[:, 3:]
                vertices_indices = []
                for i in range(comp_vertices.shape[0]):
                    v = comp_vertices[i].tostring()
                    if v in vertices_dict:
                        vertices_indices.append(vertices_dict[v][0])
                    else:
                        vertices_indices.append(vertex_idx)
                        vertices_dict[v] = (vertex_idx, comp_vertices[i])
                        vertex_idx += 1

                # Correctly transform the normals with the inverse transpose
                comp_normals = (np.linalg.inv(transform).T @ np.concatenate((mesh.normals, np.ones_like(mesh.normals[:, 0:1])), axis=1).T).T
                comp_normals = comp_normals[:, :3] / np.linalg.norm(comp_normals[:, :3], axis=1, keepdims=True).clip(1e-5)
                normals_indices = []
                for i in range(comp_normals.shape[0]):
                    n = comp_normals[i].tostring()
                    if n in normals_dict:
                        normals_indices.append(normals_dict[n][0])
                    else:
                        normals_indices.append(normal_idx)
                        normals_dict[n] = (normal_idx, comp_normals[i])
                        normal_idx += 1

                uvs_indices = []
                for i in range(mesh.texcoords.shape[0]):
                    uv = mesh.texcoords[i].tostring()
                    if uv in uvs_dict:
                        uvs_indices.append(uvs_dict[uv][0])
                    else:
                        uvs_indices.append(uvs_idx)
                        uvs_dict[uv] = (uvs_idx, mesh.texcoords[i])
                        uvs_idx += 1

                # Increment triangle indices with index offset, use correct vertex idx for reused vertices
                comp_triangles = mesh.triangles
                n_triangles = len(comp_triangles)
                for tri in comp_triangles:
                    v_idx = tuple(vertices_indices[idx] for idx in tri[0])
                    uv_idx = tuple(uvs_indices[idx] for idx in tri[1])
                    n_idx = tuple(normals_indices[idx] for idx in tri[2])
                    triangles += [(v_idx, uv_idx, n_idx)]

                # Create submesh or take over submeshes for each component
                if len(mesh.sub_meshes_raw) > 0:
                    for sub_mesh_texture in mesh.sub_meshes_raw:
                        sub_mesh, texture, normal_map = sub_mesh_texture
                        triangle_idx = []
                        for f_i in sub_mesh.faces:
                            start = mesh._faces_triangulation[f_i]
                            end = n_triangles if f_i >= len(mesh._faces_triangulation) - 1 else mesh._faces_triangulation[f_i + 1]
                            triangle_idx += list(range(start + tri_offset, end + tri_offset))
                        sub_mesh.faces = triangle_idx
                        if texture in sub_mesh_dict:
                            sub_mesh_dict[texture].append(sub_mesh)
                        else:
                            sub_mesh_dict[texture] = [sub_mesh]
                # else:
                #     sub_mesh = SubMesh(faces=(np.arange(n_triangles) + tri_offset).tolist())
                #     texture = mesh.texture
                #     if texture in sub_mesh_dict:
                #         sub_mesh_dict[texture].append(sub_mesh)
                #     else:
                #         sub_mesh_dict[texture] = [sub_mesh]
                # Offset for triangle references
                tri_offset += n_triangles
        # Create mesh
        _, vertices = zip(*list(vertices_dict.values()))
        _, normals = zip(*list(normals_dict.values()))
        _, uvs = zip(*list(uvs_dict.values()))
        mesh = Mesh(
            name=self.name,
            vertices=np.vstack(vertices).tolist(),
            colors=[Palette.GreenA.as_color()],
            normals=np.vstack(normals).tolist(),
            uvs=np.vstack(uvs).tolist(),
            faces=triangles
        )

        textures_sub_meshes = list(sub_mesh_dict.items())
        if len(textures_sub_meshes) > 0:
            texture, sub_meshes = textures_sub_meshes[0]
            sub_mesh = SubMesh.from_sub_meshes(sub_meshes, name=texture)
            mesh.update_sub_mesh(0, sub_mesh, texture=texture)
        if len(textures_sub_meshes) > 1:
            for i in range(1, len(textures_sub_meshes)):
                texture, sub_meshes = textures_sub_meshes[i]
                sub_mesh = SubMesh.from_sub_meshes(sub_meshes, name=texture)
                mesh.append_sub_mesh(sub_mesh, texture=texture)
        return mesh

    @property
    def meshes(self):
        return [(comp.mesh, self.transform_of(comp)) for comp in self._components]



import abc
from dataclasses import dataclass

import numpy as np

from .. import gl

glsl_types = {
    'float': gl.GL_FLOAT,
    'vec2': gl.GL_FLOAT_VEC2,
    'vec3': gl.GL_FLOAT_VEC3,
    'vec4': gl.GL_FLOAT_VEC4,
    'int': gl.GL_INT,
    'ivec2': gl.GL_INT_VEC2,
    'ivec3': gl.GL_INT_VEC3,
    'ivec4': gl.GL_INT_VEC4,
    'bool': gl.GL_BOOL,
    'bvec2': gl.GL_BOOL_VEC2,
    'bvec3': gl.GL_BOOL_VEC3,
    'bvec4': gl.GL_BOOL_VEC4,
    'mat2': gl.GL_FLOAT_MAT2,
    'mat3': gl.GL_FLOAT_MAT3,
    'mat4': gl.GL_FLOAT_MAT4,
    'sampler1D': gl.GL_SAMPLER_1D,
    'sampler2D': gl.GL_SAMPLER_2D,
    'samplerCube': gl.GL_SAMPLER_CUBE,
}

gl_type_info = {
    gl.GL_FLOAT: (1, gl.GL_FLOAT, np.float32),
    gl.GL_FLOAT_VEC2: (2, gl.GL_FLOAT, np.float32),
    gl.GL_FLOAT_VEC3: (3, gl.GL_FLOAT, np.float32),
    gl.GL_FLOAT_VEC4: (4, gl.GL_FLOAT, np.float32),
    gl.GL_INT: (1, gl.GL_INT, np.int32),
    gl.GL_INT_VEC2: (2, gl.GL_INT, np.int32),
    gl.GL_INT_VEC3: (3, gl.GL_INT, np.int32),
    gl.GL_INT_VEC4: (4, gl.GL_INT, np.int32),
    gl.GL_BOOL: (1, gl.GL_BOOL, np.bool),
    gl.GL_BOOL_VEC2: (2, gl.GL_BOOL, np.bool),
    gl.GL_BOOL_VEC3: (3, gl.GL_BOOL, np.bool),
    gl.GL_BOOL_VEC4: (4, gl.GL_BOOL, np.bool),
    gl.GL_FLOAT_MAT2: (4, gl.GL_FLOAT, np.float32),
    gl.GL_FLOAT_MAT3: (9, gl.GL_FLOAT, np.float32),
    gl.GL_FLOAT_MAT4: (16, gl.GL_FLOAT, np.float32),
    gl.GL_SAMPLER_1D: (1, gl.GL_UNSIGNED_INT, np.uint32),
    gl.GL_SAMPLER_2D: (1, gl.GL_UNSIGNED_INT, np.uint32),
    gl.GL_SAMPLER_CUBE: (1, gl.GL_UNSIGNED_INT, np.uint32)
}

# A lookup table to specify the value of a uniform variable for shader program.
gl_set_uniform = {
    gl.GL_FLOAT: gl.glUniform1fv,
    gl.GL_FLOAT_VEC2: gl.glUniform2fv,
    gl.GL_FLOAT_VEC3: gl.glUniform3fv,
    gl.GL_FLOAT_VEC4: gl.glUniform4fv,
    gl.GL_INT: gl.glUniform1iv,
    gl.GL_INT_VEC2: gl.glUniform2iv,
    gl.GL_INT_VEC3: gl.glUniform3iv,
    gl.GL_INT_VEC4: gl.glUniform4iv,
    gl.GL_BOOL: gl.glUniform1iv,
    gl.GL_BOOL_VEC2: gl.glUniform2iv,
    gl.GL_BOOL_VEC3: gl.glUniform3iv,
    gl.GL_BOOL_VEC4: gl.glUniform4iv,
    gl.GL_FLOAT_MAT2: gl.glUniformMatrix2fv,
    gl.GL_FLOAT_MAT3: gl.glUniformMatrix3fv,
    gl.GL_FLOAT_MAT4: gl.glUniformMatrix4fv,
    gl.GL_SAMPLER_1D: gl.glUniform1iv,
    gl.GL_SAMPLER_2D: gl.glUniform1iv,
    gl.GL_SAMPLER_CUBE: gl.glUniform1iv
}

gl_get_uniform = {
    gl.GL_FLOAT: gl.glUniform1fv,
    gl.GL_FLOAT_VEC2: gl.glGetnUniformfv,
    gl.GL_FLOAT_VEC3: gl.glGetnUniformfv,
    gl.GL_FLOAT_VEC4: gl.glGetnUniformfv,
    gl.GL_INT: gl.glGetnUniformiv,
    gl.GL_INT_VEC2: gl.glGetnUniformiv,
    gl.GL_INT_VEC3: gl.glGetnUniformiv,
    gl.GL_INT_VEC4: gl.glGetnUniformiv,
    gl.GL_BOOL: gl.glGetnUniformiv,
    gl.GL_BOOL_VEC2: gl.glGetnUniformiv,
    gl.GL_BOOL_VEC3: gl.glGetnUniformiv,
    gl.GL_BOOL_VEC4: gl.glGetnUniformiv,
    gl.GL_FLOAT_MAT2: gl.glGetnUniformfv,
    gl.GL_FLOAT_MAT3: gl.glGetnUniformfv,
    gl.GL_FLOAT_MAT4: gl.glGetnUniformfv,
    gl.GL_SAMPLER_1D: gl.glGetnUniformiv,
    gl.GL_SAMPLER_2D: gl.glGetnUniformiv,
    gl.GL_SAMPLER_CUBE: gl.glGetnUniformiv
}

# gl_vertex_attrib = {
#     gl.GL_FLOAT: gl.glVertexAttrib1f,
#     gl.GL_FLOAT_VEC2: gl.glVertexAttrib2f,
#     gl.GL_FLOAT_VEC3: gl.glVertexAttrib3f,
#     gl.GL_FLOAT_VEC4: gl.glVertexAttrib3f,
#     gl.GL_INT: gl.glVertexAttribI1i,
#     gl.GL_INT_VEC2: gl.glVertexAttribI2i,
#     gl.GL_INT_VEC3: gl.glVertexAttribI3i,
#     gl.GL_INT_VEC4: gl.glVertexAttribI4i,
#     gl.GL_BOOL: gl.glVertexAttribI1ui,
#     gl.GL_BOOL_VEC2: gl.glVertexAttrib2fv,
#     gl.GL_BOOL_VEC3: gl.glUniform3iv,
#     gl.GL_BOOL_VEC4: gl.glUniform4iv,
#     gl.GL_FLOAT_MAT2: gl.glUniformMatrix2fv,
#     gl.GL_FLOAT_MAT3: gl.glUniformMatrix3fv,
#     gl.GL_FLOAT_MAT4: gl.glUniformMatrix4fv,
# }


@dataclass
class AbstractDataclass(abc.ABC):
    def __new__(cls, *args, **kwargs):
        if cls == AbstractDataclass or cls.__bases__[0] == AbstractDataclass:
            raise TypeError("Cannot instantiate an abstract class.")
        return super().__new__(cls)


@dataclass
class ShaderVariable(AbstractDataclass):
    """Interface for shader variables: attribute(in) and uniform"""
    program: gl.Constant
    name: str
    gtype: gl.Constant
    dtype: np.dtype
    size: int

    def __init__(self, program, name, gl_type: gl.Constant):
        if gl_type not in gl_type_info:
            raise TypeError("Unknown shader variable type")
        self.program = program
        self.name = name
        self.gtype = gl_type
        self.dtype = gl_type_info['gl_type']


class ShaderAttribute(ShaderVariable):
    def __init__(self, program, name, attr_type, size, location):
        super(ShaderAttribute, self).__init__(program, name)
        self.size = size
        self.location = location


class ShaderUniform(ShaderVariable):
    def __init__(self, program, name, uniform_type, gl_type, location, length, count, gl_setter, gl_getter):
        self.program = program
        self.name = name
        self.type = uniform_type
        self.location = location
        self.length = length
        self.count = count
